"""
Azure ARM Cloud Module
======================

.. versionadded:: 2016.11.0

.. versionchanged:: 2019.2.0

The Azure ARM cloud module is used to control access to Microsoft Azure Resource Manager

.. warning::

    This cloud provider will be removed from Salt in version 3007 in favor of
    the `saltext.azurerm Salt Extension
    <https://github.com/salt-extensions/saltext-azurerm>`_

:maintainer: <devops@eitr.tech>
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0rc6
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.4
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 0.30.0rc6
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 0.33.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 0.30.0rc6
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 0.30.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 0.30.0rc6
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.30.0rc6
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.32.0
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:configuration:
    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    if using Managed Service Identity authentication:
      * ``subscription_id``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``
      * HTTP base URL for a custom endpoint, such as Azure Stack. The ``/metadata/endpoints`` path will be added to the URL.

    **userdata** and **userdata_file**:
      Azure Resource Manager uses a separate VirtualMachineExtension object to pass userdata scripts to the virtual
      machine. Arbitrary shell commands can be passed via the ``userdata`` parameter, or via a file local to the Salt
      Cloud system using the ``userdata_file`` parameter. Note that the local file is not treated as a script by the
      extension, so "one-liners" probably work best. If greater functionality is desired, a web-hosted script file can
      be specified via ``userdata_file: https://raw.githubusercontent.com/account/repo/master/azure-script.py``, which
      will be executed on the system after VM creation. For Windows systems, script files ending in ``.ps1`` will be
      executed with ``powershell.exe``. The ``userdata`` parameter takes precedence over the ``userdata_file`` parameter
      when creating the custom script extension.

    **win_installer**:
      This parameter, which holds the local path to the Salt Minion installer package, is used to determine if the
      virtual machine type will be "Windows". Only set this parameter on profiles which install Windows operating systems.


Example ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/azure.conf`` configuration:

.. code-block:: yaml

    my-azure-config with username and password:
      driver: azurearm
      subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
      username: larry
      password: 123pass

    Or my-azure-config with service principal:
      driver: azurearm
      subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
      tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
      client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
      secret: XXXXXXXXXXXXXXXXXXXXXXXX
      cloud_environment: AZURE_US_GOV_CLOUD

      The Service Principal can be created with the new Azure CLI (https://github.com/Azure/azure-cli) with:
      az ad sp create-for-rbac -n "http://<yourappname>" --role <role> --scopes <scope>
      For example, this creates a service principal with 'owner' role for the whole subscription:
      az ad sp create-for-rbac -n "http://mysaltapp" --role owner --scopes /subscriptions/3287abc8-f98a-c678-3bde-326766fd3617

      *Note: review the details of Service Principals. Owner role is more than you normally need, and you can restrict
      scope to a resource group or individual resources.
"""

import importlib
import logging
import os
import os.path
import pprint
import string
import time
from functools import wraps
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool

import salt.cache
import salt.config as config
import salt.loader
import salt.utils.azurearm
import salt.utils.cloud
import salt.utils.files
import salt.utils.stringutils
import salt.utils.yaml
import salt.version
from salt.exceptions import (
    SaltCloudConfigError,
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudSystemExit,
)

HAS_LIBS = False
try:
    import azure.mgmt.compute.models as compute_models
    import azure.mgmt.network.models as network_models
    from azure.storage.blob.blockblobservice import BlockBlobService
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__virtualname__ = "azurearm"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Check for Azure configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return (
            False,
            "The following dependencies are required to use the AzureARM driver: "
            "Microsoft Azure SDK for Python >= 2.0rc6, "
            "Microsoft Azure Storage SDK for Python >= 0.32, "
            "MS REST Azure (msrestazure) >= 0.4",
        )

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def _deprecation_message(function):
    """
    Decorator wrapper to warn about msazure deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Chlorine",
            "This cloud provider will be removed from Salt in version 3007 due to "
            "the deprecation of the 'Classic' API for Azure. Please migrate to "
            "Azure Resource Manager by March 1, 2023 "
            "(https://docs.microsoft.com/en-us/azure/virtual-machines/classic-vm-deprecation)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def get_api_versions(call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    Get a resource type api versions
    """
    if kwargs is None:
        kwargs = {}

    if "resource_provider" not in kwargs:
        raise SaltCloudSystemExit("A resource_provider must be specified")

    if "resource_type" not in kwargs:
        raise SaltCloudSystemExit("A resource_type must be specified")

    api_versions = []

    try:
        resconn = get_conn(client_type="resource")
        provider_query = resconn.providers.get(
            resource_provider_namespace=kwargs["resource_provider"]
        )

        for resource in provider_query.resource_types:
            if str(resource.resource_type) == kwargs["resource_type"]:
                resource_dict = resource.as_dict()
                api_versions = resource_dict["api_versions"]
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("resource", exc.message)

    return api_versions


@_deprecation_message
def get_resource_by_id(resource_id, api_version, extract_value=None):
    """
    Get an AzureARM resource by id
    """
    ret = {}

    try:
        resconn = get_conn(client_type="resource")
        resource_query = resconn.resources.get_by_id(
            resource_id=resource_id, api_version=api_version
        )
        resource_dict = resource_query.as_dict()
        if extract_value is not None:
            ret = resource_dict[extract_value]
        else:
            ret = resource_dict
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("resource", exc.message)
        ret = {"Error": exc.message}

    return ret


def get_configured_provider():
    """
    Return the first configured provider instance.
    """
    key_combos = [
        ("subscription_id", "tenant", "client_id", "secret"),
        ("subscription_id", "username", "password"),
        ("subscription_id",),
    ]

    for combo in key_combos:
        provider = config.is_provider_configured(
            __opts__,
            _get_active_provider_name() or __virtualname__,
            combo,
        )

        if provider:
            return provider

    return provider


@_deprecation_message
def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(__virtualname__, {"azurearm": HAS_LIBS})


@_deprecation_message
def get_conn(client_type):
    """
    Return a connection object for a client type.
    """
    conn_kwargs = {}

    conn_kwargs["subscription_id"] = salt.utils.stringutils.to_str(
        config.get_cloud_config_value(
            "subscription_id", get_configured_provider(), __opts__, search_global=False
        )
    )

    cloud_env = config.get_cloud_config_value(
        "cloud_environment", get_configured_provider(), __opts__, search_global=False
    )

    if cloud_env is not None:
        conn_kwargs["cloud_environment"] = cloud_env

    tenant = config.get_cloud_config_value(
        "tenant", get_configured_provider(), __opts__, search_global=False
    )

    if tenant is not None:
        client_id = config.get_cloud_config_value(
            "client_id", get_configured_provider(), __opts__, search_global=False
        )
        secret = config.get_cloud_config_value(
            "secret", get_configured_provider(), __opts__, search_global=False
        )
        conn_kwargs.update({"client_id": client_id, "secret": secret, "tenant": tenant})

    username = config.get_cloud_config_value(
        "username", get_configured_provider(), __opts__, search_global=False
    )

    if username:
        password = config.get_cloud_config_value(
            "password", get_configured_provider(), __opts__, search_global=False
        )
        conn_kwargs.update({"username": username, "password": password})

    client = salt.utils.azurearm.get_client(client_type=client_type, **conn_kwargs)

    return client


@_deprecation_message
def get_location(call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    Return the location that is configured for this provider
    """
    if not kwargs:
        kwargs = {}
    vm_dict = get_configured_provider()
    vm_dict.update(kwargs)
    return config.get_cloud_config_value(
        "location", vm_dict, __opts__, search_global=False
    )


@_deprecation_message
def avail_locations(call=None):
    """
    Return a dict of all available regions.
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    ret = {}
    ret["locations"] = []

    try:
        resconn = get_conn(client_type="resource")
        provider_query = resconn.providers.get(
            resource_provider_namespace="Microsoft.Compute"
        )
        locations = []
        for resource in provider_query.resource_types:
            if str(resource.resource_type) == "virtualMachines":
                resource_dict = resource.as_dict()
                locations = resource_dict["locations"]
        for location in locations:
            lowercase = location.lower().replace(" ", "")
            ret["locations"].append(lowercase)
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("resource", exc.message)
        ret = {"Error": exc.message}

    return ret


@_deprecation_message
def avail_images(call=None):
    """
    Return a dict of all available images on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )
    compconn = get_conn(client_type="compute")
    region = get_location()
    publishers = []
    ret = {}

    def _get_publisher_images(publisher):
        """
        Get all images from a specific publisher
        """
        data = {}
        try:
            offers = compconn.virtual_machine_images.list_offers(
                location=region,
                publisher_name=publisher,
            )
            for offer_obj in offers:
                offer = offer_obj.as_dict()
                skus = compconn.virtual_machine_images.list_skus(
                    location=region,
                    publisher_name=publisher,
                    offer=offer["name"],
                )
                for sku_obj in skus:
                    sku = sku_obj.as_dict()
                    results = compconn.virtual_machine_images.list(
                        location=region,
                        publisher_name=publisher,
                        offer=offer["name"],
                        skus=sku["name"],
                    )
                    for version_obj in results:
                        version = version_obj.as_dict()
                        name = "|".join(
                            (
                                publisher,
                                offer["name"],
                                sku["name"],
                                version["name"],
                            )
                        )
                        data[name] = {
                            "publisher": publisher,
                            "offer": offer["name"],
                            "sku": sku["name"],
                            "version": version["name"],
                        }
        except CloudError as exc:
            salt.utils.azurearm.log_cloud_error("compute", exc.message)
            data = {publisher: exc.message}

        return data

    try:
        publishers_query = compconn.virtual_machine_images.list_publishers(
            location=region
        )
        for publisher_obj in publishers_query:
            publisher = publisher_obj.as_dict()
            publishers.append(publisher["name"])
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("compute", exc.message)

    pool = ThreadPool(cpu_count() * 6)
    results = pool.map_async(_get_publisher_images, publishers)
    results.wait()

    ret = {k: v for result in results.get() for k, v in result.items()}

    return ret


@_deprecation_message
def avail_sizes(call=None):
    """
    Return a list of sizes available from the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    compconn = get_conn(client_type="compute")

    ret = {}
    location = get_location()

    try:
        sizes = compconn.virtual_machine_sizes.list(location=location)
        for size_obj in sizes:
            size = size_obj.as_dict()
            ret[size["name"]] = size
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("compute", exc.message)
        ret = {"Error": exc.message}

    return ret


@_deprecation_message
def list_nodes(call=None):
    """
    List VMs on this Azure account
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    ret = {}

    nodes = list_nodes_full()
    for node in nodes:
        ret[node] = {"name": node}
        for prop in ("id", "image", "size", "state", "private_ips", "public_ips"):
            ret[node][prop] = nodes[node].get(prop)
    return ret


@_deprecation_message
def list_nodes_full(call=None):
    """
    List all VMs on the subscription with full information
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    netapi_versions = get_api_versions(
        kwargs={
            "resource_provider": "Microsoft.Network",
            "resource_type": "networkInterfaces",
        }
    )
    netapi_version = netapi_versions[0]
    compconn = get_conn(client_type="compute")

    ret = {}

    def _get_node_info(node):
        """
        Get node info.
        """
        node_ret = {}
        node["id"] = node["vm_id"]
        node["size"] = node["hardware_profile"]["vm_size"]
        node["state"] = node["provisioning_state"]
        node["public_ips"] = []
        node["private_ips"] = []
        node_ret[node["name"]] = node
        try:
            image_ref = node["storage_profile"]["image_reference"]
            node["image"] = "|".join(
                [
                    image_ref["publisher"],
                    image_ref["offer"],
                    image_ref["sku"],
                    image_ref["version"],
                ]
            )
        except (TypeError, KeyError):
            try:
                node["image"] = node["storage_profile"]["os_disk"]["image"]["uri"]
            except (TypeError, KeyError):
                node["image"] = (
                    node.get("storage_profile", {}).get("image_reference", {}).get("id")
                )
        try:
            netifaces = node["network_profile"]["network_interfaces"]
            for index, netiface in enumerate(netifaces):
                netiface_name = get_resource_by_id(
                    netiface["id"], netapi_version, "name"
                )
                netiface, pubips, privips = _get_network_interface(
                    netiface_name, node["resource_group"]
                )
                node["network_profile"]["network_interfaces"][index].update(netiface)
                node["public_ips"].extend(pubips)
                node["private_ips"].extend(privips)
        except Exception:  # pylint: disable=broad-except
            pass

        node_ret[node["name"]] = node

        return node_ret

    for group in list_resource_groups():
        nodes = []
        nodes_query = compconn.virtual_machines.list(resource_group_name=group)
        for node_obj in nodes_query:
            node = node_obj.as_dict()
            node["resource_group"] = group
            nodes.append(node)

        pool = ThreadPool(cpu_count() * 6)
        results = pool.map_async(_get_node_info, nodes)
        results.wait()

        group_ret = {k: v for result in results.get() for k, v in result.items()}
        ret.update(group_ret)

    return ret


@_deprecation_message
def list_resource_groups(call=None):
    """
    List resource groups associated with the subscription
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_hosted_services function must be called with -f or --function"
        )

    resconn = get_conn(client_type="resource")
    ret = {}
    try:
        groups = resconn.resource_groups.list()

        for group_obj in groups:
            group = group_obj.as_dict()
            ret[group["name"]] = group
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("resource", exc.message)
        ret = {"Error": exc.message}

    return ret


@_deprecation_message
def show_instance(name, call=None):
    """
    Show the details from AzureARM concerning an instance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )
    try:
        node = list_nodes_full("function")[name]
    except KeyError:
        log.debug("Failed to get data for node '%s'", name)
        node = {}

    __utils__["cloud.cache_node"](node, _get_active_provider_name(), __opts__)

    return node


@_deprecation_message
def delete_interface(call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    Delete a network interface.
    """
    if kwargs is None:
        kwargs = {}

    netconn = get_conn(client_type="network")

    if kwargs.get("resource_group") is None:
        kwargs["resource_group"] = config.get_cloud_config_value(
            "resource_group", {}, __opts__, search_global=True
        )

    ips = []
    iface = netconn.network_interfaces.get(
        kwargs["resource_group"],
        kwargs["iface_name"],
    )
    iface_name = iface.name
    for ip_ in iface.ip_configurations:
        ips.append(ip_.name)

    poller = netconn.network_interfaces.delete(
        kwargs["resource_group"],
        kwargs["iface_name"],
    )
    poller.wait()

    for ip_ in ips:
        poller = netconn.public_ip_addresses.delete(kwargs["resource_group"], ip_)
        poller.wait()

    return {iface_name: ips}


def _get_public_ip(name, resource_group):
    """
    Get the public ip address details by name.
    """
    netconn = get_conn(client_type="network")
    try:
        pubip_query = netconn.public_ip_addresses.get(
            resource_group_name=resource_group, public_ip_address_name=name
        )
        pubip = pubip_query.as_dict()
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("network", exc.message)
        pubip = {"error": exc.message}

    return pubip


def _get_network_interface(name, resource_group):
    """
    Get a network interface.
    """
    public_ips = []
    private_ips = []
    netapi_versions = get_api_versions(
        kwargs={
            "resource_provider": "Microsoft.Network",
            "resource_type": "publicIPAddresses",
        }
    )
    netapi_version = netapi_versions[0]
    netconn = get_conn(client_type="network")
    netiface_query = netconn.network_interfaces.get(
        resource_group_name=resource_group, network_interface_name=name
    )

    netiface = netiface_query.as_dict()
    for index, ip_config in enumerate(netiface["ip_configurations"]):
        if ip_config.get("private_ip_address") is not None:
            private_ips.append(ip_config["private_ip_address"])
        if "id" in ip_config.get("public_ip_address", {}):
            public_ip_name = get_resource_by_id(
                ip_config["public_ip_address"]["id"], netapi_version, "name"
            )
            public_ip = _get_public_ip(public_ip_name, resource_group)
            public_ips.append(public_ip["ip_address"])
            netiface["ip_configurations"][index]["public_ip_address"].update(public_ip)

    return netiface, public_ips, private_ips


@_deprecation_message
def create_network_interface(call=None, kwargs=None):
    """
    Create a network interface.
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The create_network_interface action must be called with -a or --action."
        )

    # pylint: disable=invalid-name
    IPAllocationMethod = getattr(network_models, "IPAllocationMethod")
    # pylint: disable=invalid-name
    NetworkInterface = getattr(network_models, "NetworkInterface")
    # pylint: disable=invalid-name
    NetworkInterfaceIPConfiguration = getattr(
        network_models, "NetworkInterfaceIPConfiguration"
    )
    # pylint: disable=invalid-name
    PublicIPAddress = getattr(network_models, "PublicIPAddress")

    if not isinstance(kwargs, dict):
        kwargs = {}

    vm_ = kwargs
    netconn = get_conn(client_type="network")

    if kwargs.get("location") is None:
        kwargs["location"] = get_location()

    if kwargs.get("network") is None:
        kwargs["network"] = config.get_cloud_config_value(
            "network", vm_, __opts__, search_global=False
        )

    if kwargs.get("subnet") is None:
        kwargs["subnet"] = config.get_cloud_config_value(
            "subnet", vm_, __opts__, search_global=False
        )

    if kwargs.get("network_resource_group") is None:
        kwargs["network_resource_group"] = config.get_cloud_config_value(
            "resource_group", vm_, __opts__, search_global=False
        )

    if kwargs.get("iface_name") is None:
        kwargs["iface_name"] = "{}-iface0".format(vm_["name"])

    try:
        subnet_obj = netconn.subnets.get(
            resource_group_name=kwargs["network_resource_group"],
            virtual_network_name=kwargs["network"],
            subnet_name=kwargs["subnet"],
        )
    except CloudError as exc:
        raise SaltCloudSystemExit(
            '{} (Resource Group: "{}", VNET: "{}", Subnet: "{}")'.format(
                exc.message,
                kwargs["network_resource_group"],
                kwargs["network"],
                kwargs["subnet"],
            )
        )

    ip_kwargs = {}
    ip_configurations = None

    if "load_balancer_backend_address_pools" in kwargs:
        pool_dicts = kwargs["load_balancer_backend_address_pools"]
        if isinstance(pool_dicts, dict):
            pool_ids = []
            for load_bal, be_pools in pool_dicts.items():
                for pool in be_pools:
                    try:
                        lbbep_data = netconn.load_balancer_backend_address_pools.get(
                            kwargs["resource_group"],
                            load_bal,
                            pool,
                        )
                        pool_ids.append({"id": lbbep_data.as_dict()["id"]})
                    except CloudError as exc:
                        log.error("There was a cloud error: %s", str(exc))
                    except KeyError as exc:
                        log.error(
                            "There was an error getting the Backend Pool ID: %s",
                            str(exc),
                        )
            ip_kwargs["load_balancer_backend_address_pools"] = pool_ids

    if "private_ip_address" in kwargs.keys():
        ip_kwargs["private_ip_address"] = kwargs["private_ip_address"]
        ip_kwargs["private_ip_allocation_method"] = IPAllocationMethod.static
    else:
        ip_kwargs["private_ip_allocation_method"] = IPAllocationMethod.dynamic

    if kwargs.get("allocate_public_ip") is True:
        pub_ip_name = "{}-ip".format(kwargs["iface_name"])
        poller = netconn.public_ip_addresses.create_or_update(
            resource_group_name=kwargs["resource_group"],
            public_ip_address_name=pub_ip_name,
            parameters=PublicIPAddress(
                location=kwargs["location"],
                public_ip_allocation_method=IPAllocationMethod.static,
            ),
        )
        count = 0
        poller.wait()
        while True:
            try:
                pub_ip_data = netconn.public_ip_addresses.get(
                    kwargs["resource_group"],
                    pub_ip_name,
                )
                if pub_ip_data.ip_address:  # pylint: disable=no-member
                    ip_kwargs["public_ip_address"] = PublicIPAddress(
                        id=str(pub_ip_data.id),  # pylint: disable=no-member
                    )
                    ip_configurations = [
                        NetworkInterfaceIPConfiguration(
                            name="{}-ip".format(kwargs["iface_name"]),
                            subnet=subnet_obj,
                            **ip_kwargs
                        )
                    ]
                    break
            except CloudError as exc:
                log.error("There was a cloud error: %s", exc)
            count += 1
            if count > 120:
                raise ValueError("Timed out waiting for public IP Address.")
            time.sleep(5)
    else:
        priv_ip_name = "{}-ip".format(kwargs["iface_name"])
        ip_configurations = [
            NetworkInterfaceIPConfiguration(
                name=priv_ip_name, subnet=subnet_obj, **ip_kwargs
            )
        ]

    network_security_group = None
    if kwargs.get("security_group") is not None:
        network_security_group = netconn.network_security_groups.get(
            resource_group_name=kwargs["resource_group"],
            network_security_group_name=kwargs["security_group"],
        )

    iface_params = NetworkInterface(
        location=kwargs["location"],
        network_security_group=network_security_group,
        ip_configurations=ip_configurations,
    )

    poller = netconn.network_interfaces.create_or_update(
        kwargs["resource_group"], kwargs["iface_name"], iface_params
    )
    try:
        poller.wait()
    except Exception as exc:  # pylint: disable=broad-except
        log.warning(
            "Network interface creation could not be polled. "
            "It is likely that we are reusing an existing interface. (%s)",
            exc,
        )

    count = 0
    while True:
        try:
            return _get_network_interface(
                kwargs["iface_name"], kwargs["resource_group"]
            )
        except CloudError:
            count += 1
            if count > 120:
                raise ValueError("Timed out waiting for operation to complete.")
            time.sleep(5)


def request_instance(vm_, kwargs=None):
    """
    Request a VM from Azure.
    """
    compconn = get_conn(client_type="compute")

    # pylint: disable=invalid-name
    CachingTypes = getattr(compute_models, "CachingTypes")
    # pylint: disable=invalid-name
    DataDisk = getattr(compute_models, "DataDisk")
    # pylint: disable=invalid-name
    DiskCreateOptionTypes = getattr(compute_models, "DiskCreateOptionTypes")
    # pylint: disable=invalid-name
    HardwareProfile = getattr(compute_models, "HardwareProfile")
    # pylint: disable=invalid-name
    ImageReference = getattr(compute_models, "ImageReference")
    # pylint: disable=invalid-name
    LinuxConfiguration = getattr(compute_models, "LinuxConfiguration")
    # pylint: disable=invalid-name
    SshConfiguration = getattr(compute_models, "SshConfiguration")
    # pylint: disable=invalid-name
    SshPublicKey = getattr(compute_models, "SshPublicKey")
    # pylint: disable=invalid-name
    NetworkInterfaceReference = getattr(compute_models, "NetworkInterfaceReference")
    # pylint: disable=invalid-name
    NetworkProfile = getattr(compute_models, "NetworkProfile")
    # pylint: disable=invalid-name
    OSDisk = getattr(compute_models, "OSDisk")
    # pylint: disable=invalid-name
    OSProfile = getattr(compute_models, "OSProfile")
    # pylint: disable=invalid-name
    StorageProfile = getattr(compute_models, "StorageProfile")
    # pylint: disable=invalid-name
    VirtualHardDisk = getattr(compute_models, "VirtualHardDisk")
    # pylint: disable=invalid-name
    VirtualMachine = getattr(compute_models, "VirtualMachine")
    # pylint: disable=invalid-name
    VirtualMachineSizeTypes = getattr(compute_models, "VirtualMachineSizeTypes")

    subscription_id = config.get_cloud_config_value(
        "subscription_id", get_configured_provider(), __opts__, search_global=False
    )

    if vm_.get("driver") is None:
        vm_["driver"] = "azurearm"

    if vm_.get("location") is None:
        vm_["location"] = get_location()

    if vm_.get("resource_group") is None:
        vm_["resource_group"] = config.get_cloud_config_value(
            "resource_group", vm_, __opts__, search_global=True
        )

    if vm_.get("name") is None:
        vm_["name"] = config.get_cloud_config_value(
            "name", vm_, __opts__, search_global=True
        )

    # pylint: disable=unused-variable
    iface_data, public_ips, private_ips = create_network_interface(
        call="action", kwargs=vm_
    )
    vm_["iface_id"] = iface_data["id"]

    disk_name = "{}-vol0".format(vm_["name"])

    vm_username = config.get_cloud_config_value(
        "ssh_username",
        vm_,
        __opts__,
        search_global=True,
        default=config.get_cloud_config_value(
            "win_username", vm_, __opts__, search_global=True
        ),
    )

    ssh_publickeyfile_contents = None
    ssh_publickeyfile = config.get_cloud_config_value(
        "ssh_publickeyfile", vm_, __opts__, search_global=False, default=None
    )
    if ssh_publickeyfile is not None:
        try:
            with salt.utils.files.fopen(ssh_publickeyfile, "r") as spkc_:
                ssh_publickeyfile_contents = spkc_.read()
        except Exception as exc:  # pylint: disable=broad-except
            raise SaltCloudConfigError(
                "Failed to read ssh publickey file '{}': {}".format(
                    ssh_publickeyfile, exc.args[-1]
                )
            )

    disable_password_authentication = config.get_cloud_config_value(
        "disable_password_authentication",
        vm_,
        __opts__,
        search_global=False,
        default=False,
    )

    os_kwargs = {}
    win_installer = config.get_cloud_config_value(
        "win_installer", vm_, __opts__, search_global=True
    )
    if not win_installer and ssh_publickeyfile_contents is not None:
        sshpublickey = SshPublicKey(
            key_data=ssh_publickeyfile_contents,
            path="/home/{}/.ssh/authorized_keys".format(vm_username),
        )
        sshconfiguration = SshConfiguration(
            public_keys=[sshpublickey],
        )
        linuxconfiguration = LinuxConfiguration(
            disable_password_authentication=disable_password_authentication,
            ssh=sshconfiguration,
        )
        os_kwargs["linux_configuration"] = linuxconfiguration
        vm_password = None
    else:
        vm_password = salt.utils.stringutils.to_str(
            config.get_cloud_config_value(
                "ssh_password",
                vm_,
                __opts__,
                search_global=True,
                default=config.get_cloud_config_value(
                    "win_password", vm_, __opts__, search_global=True
                ),
            )
        )

    if win_installer or (
        vm_password is not None and not disable_password_authentication
    ):
        if not isinstance(vm_password, str):
            raise SaltCloudSystemExit("The admin password must be a string.")
        if len(vm_password) < 8 or len(vm_password) > 123:
            raise SaltCloudSystemExit(
                "The admin password must be between 8-123 characters long."
            )
        complexity = 0
        if any(char.isdigit() for char in vm_password):
            complexity += 1
        if any(char.isupper() for char in vm_password):
            complexity += 1
        if any(char.islower() for char in vm_password):
            complexity += 1
        if any(char in string.punctuation for char in vm_password):
            complexity += 1
        if complexity < 3:
            raise SaltCloudSystemExit(
                "The admin password must contain at least 3 of the following types: "
                "upper, lower, digits, special characters"
            )
        os_kwargs["admin_password"] = vm_password

    availability_set = config.get_cloud_config_value(
        "availability_set", vm_, __opts__, search_global=False, default=None
    )
    if availability_set is not None and isinstance(availability_set, str):
        availability_set = {
            "id": "/subscriptions/{}/resourceGroups/{}/providers/Microsoft.Compute/availabilitySets/{}".format(
                subscription_id, vm_["resource_group"], availability_set
            )
        }
    else:
        availability_set = None

    cloud_env = _get_cloud_environment()

    storage_endpoint_suffix = cloud_env.suffixes.storage_endpoint

    if isinstance(vm_.get("volumes"), str):
        volumes = salt.utils.yaml.safe_load(vm_["volumes"])
    else:
        volumes = vm_.get("volumes")

    data_disks = None
    if isinstance(volumes, list):
        data_disks = []
    else:
        volumes = []

    lun = 0
    luns = []
    for volume in volumes:
        if isinstance(volume, str):
            volume = {"name": volume}

        volume.setdefault(
            "name",
            volume.get(
                "name",
                volume.get("name", "{}-datadisk{}".format(vm_["name"], str(lun))),
            ),
        )

        volume.setdefault(
            "disk_size_gb",
            volume.get("logical_disk_size_in_gb", volume.get("size", 100)),
        )
        # Old kwarg was host_caching, new name is caching
        volume.setdefault("caching", volume.get("host_caching", "ReadOnly"))
        while lun in luns:
            lun += 1
            if lun > 15:
                log.error("Maximum lun count has been reached")
                break
        volume.setdefault("lun", lun)
        lun += 1
        # The default vhd is {vm_name}-datadisk{lun}.vhd
        if "media_link" in volume:
            volume["vhd"] = VirtualHardDisk(uri=volume["media_link"])
            del volume["media_link"]
        elif volume.get("vhd") == "unmanaged":
            volume["vhd"] = VirtualHardDisk(
                uri="https://{}.blob.{}/vhds/{}-datadisk{}.vhd".format(
                    vm_["storage_account"],
                    storage_endpoint_suffix,
                    vm_["name"],
                    volume["lun"],
                ),
            )
        elif "vhd" in volume:
            volume["vhd"] = VirtualHardDisk(uri=volume["vhd"])

        if "image" in volume:
            volume["create_option"] = "from_image"
        elif "attach" in volume:
            volume["create_option"] = "attach"
        else:
            volume["create_option"] = "empty"
        data_disks.append(DataDisk(**volume))

    img_ref = None
    if vm_["image"].startswith("http") or vm_.get("vhd") == "unmanaged":
        if vm_["image"].startswith("http"):
            source_image = VirtualHardDisk(uri=vm_["image"])
        else:
            source_image = None
            if "|" in vm_["image"]:
                img_pub, img_off, img_sku, img_ver = vm_["image"].split("|")
                img_ref = ImageReference(
                    publisher=img_pub,
                    offer=img_off,
                    sku=img_sku,
                    version=img_ver,
                )
            elif vm_["image"].startswith("/subscriptions"):
                img_ref = ImageReference(id=vm_["image"])
        if win_installer:
            os_type = "Windows"
        else:
            os_type = "Linux"
        os_disk = OSDisk(
            caching=CachingTypes.none,
            create_option=DiskCreateOptionTypes.from_image,
            name=disk_name,
            vhd=VirtualHardDisk(
                uri="https://{}.blob.{}/vhds/{}.vhd".format(
                    vm_["storage_account"],
                    storage_endpoint_suffix,
                    disk_name,
                ),
            ),
            os_type=os_type,
            image=source_image,
            disk_size_gb=vm_.get("os_disk_size_gb"),
        )
    else:
        source_image = None
        os_type = None
        os_disk = OSDisk(
            create_option=DiskCreateOptionTypes.from_image,
            disk_size_gb=vm_.get("os_disk_size_gb"),
        )
        if "|" in vm_["image"]:
            img_pub, img_off, img_sku, img_ver = vm_["image"].split("|")
            img_ref = ImageReference(
                publisher=img_pub,
                offer=img_off,
                sku=img_sku,
                version=img_ver,
            )
        elif vm_["image"].startswith("/subscriptions"):
            img_ref = ImageReference(id=vm_["image"])

    userdata_file = config.get_cloud_config_value(
        "userdata_file", vm_, __opts__, search_global=False, default=None
    )
    userdata = config.get_cloud_config_value(
        "userdata", vm_, __opts__, search_global=False, default=None
    )
    userdata_template = config.get_cloud_config_value(
        "userdata_template", vm_, __opts__, search_global=False, default=None
    )

    if userdata_file:
        if os.path.exists(userdata_file):
            with salt.utils.files.fopen(userdata_file, "r") as fh_:
                userdata = fh_.read()

    if userdata and userdata_template:
        userdata_sendkeys = config.get_cloud_config_value(
            "userdata_sendkeys", vm_, __opts__, search_global=False, default=None
        )
        if userdata_sendkeys:
            vm_["priv_key"], vm_["pub_key"] = salt.utils.cloud.gen_keys(
                config.get_cloud_config_value("keysize", vm_, __opts__)
            )

            key_id = vm_.get("name")
            if "append_domain" in vm_:
                key_id = ".".join([key_id, vm_["append_domain"]])

            salt.utils.cloud.accept_key(__opts__["pki_dir"], vm_["pub_key"], key_id)

        userdata = salt.utils.cloud.userdata_template(__opts__, vm_, userdata)

    custom_extension = None
    if userdata is not None or userdata_file is not None:
        try:
            if win_installer:
                publisher = "Microsoft.Compute"
                virtual_machine_extension_type = "CustomScriptExtension"
                type_handler_version = "1.8"
                if userdata_file and userdata_file.endswith(".ps1"):
                    command_prefix = "powershell -ExecutionPolicy Unrestricted -File "
                else:
                    command_prefix = ""
            else:
                publisher = "Microsoft.Azure.Extensions"
                virtual_machine_extension_type = "CustomScript"
                type_handler_version = "2.0"
                command_prefix = ""

            settings = {}
            if userdata:
                settings["commandToExecute"] = userdata
            elif userdata_file.startswith("http"):
                settings["fileUris"] = [userdata_file]
                settings["commandToExecute"] = (
                    command_prefix
                    + "./"
                    + userdata_file[userdata_file.rfind("/") + 1 :]
                )

            custom_extension = {
                "resource_group": vm_["resource_group"],
                "virtual_machine_name": vm_["name"],
                "extension_name": vm_["name"] + "_custom_userdata_script",
                "location": vm_["location"],
                "publisher": publisher,
                "virtual_machine_extension_type": virtual_machine_extension_type,
                "type_handler_version": type_handler_version,
                "auto_upgrade_minor_version": True,
                "settings": settings,
                "protected_settings": None,
            }
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Failed to encode userdata: %s", exc)

    params = VirtualMachine(
        location=vm_["location"],
        plan=None,
        hardware_profile=HardwareProfile(
            vm_size=getattr(VirtualMachineSizeTypes, vm_["size"].lower(), kwargs),
        ),
        storage_profile=StorageProfile(
            os_disk=os_disk,
            data_disks=data_disks,
            image_reference=img_ref,
        ),
        os_profile=OSProfile(
            admin_username=vm_username, computer_name=vm_["name"], **os_kwargs
        ),
        network_profile=NetworkProfile(
            network_interfaces=[NetworkInterfaceReference(id=vm_["iface_id"])],
        ),
        availability_set=availability_set,
    )

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "requesting", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    try:
        vm_create = compconn.virtual_machines.create_or_update(
            resource_group_name=vm_["resource_group"],
            vm_name=vm_["name"],
            parameters=params,
        )
        vm_create.wait()
        vm_result = vm_create.result()
        vm_result = vm_result.as_dict()
        if custom_extension:
            create_or_update_vmextension(kwargs=custom_extension)
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("compute", exc.message)
        vm_result = {}

    return vm_result


@_deprecation_message
def create(vm_):
    """
    Create a single VM from a data dict.
    """
    try:
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "azurearm",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    if vm_.get("bootstrap_interface") is None:
        vm_["bootstrap_interface"] = "public"

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )
    __utils__["cloud.cachedir_index_add"](
        vm_["name"], vm_["profile"], "azurearm", vm_["driver"]
    )
    if not vm_.get("location"):
        vm_["location"] = get_location(kwargs=vm_)

    log.info("Creating Cloud VM %s in %s", vm_["name"], vm_["location"])

    vm_request = request_instance(vm_=vm_)

    if not vm_request or "error" in vm_request:
        err_message = "Error creating VM {}! ({})".format(vm_["name"], str(vm_request))
        log.error(err_message)
        raise SaltCloudSystemExit(err_message)

    def _query_node_data(name, bootstrap_interface):
        """
        Query node data.
        """
        data = show_instance(name, call="action")
        if not data:
            return False
        ip_address = None
        if bootstrap_interface == "public":
            ip_address = data["public_ips"][0]
        if bootstrap_interface == "private":
            ip_address = data["private_ips"][0]
        if ip_address is None:
            return False
        return ip_address

    try:
        data = salt.utils.cloud.wait_for_ip(
            _query_node_data,
            update_args=(
                vm_["name"],
                vm_["bootstrap_interface"],
            ),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
            interval_multiplier=config.get_cloud_config_value(
                "wait_for_ip_interval_multiplier", vm_, __opts__, default=1
            ),
        )
    except (
        SaltCloudExecutionTimeout,
        SaltCloudExecutionFailure,
        SaltCloudSystemExit,
    ) as exc:
        try:
            log.warning(exc)
        finally:
            raise SaltCloudSystemExit(str(exc))

    vm_["ssh_host"] = data
    if not vm_.get("ssh_username"):
        vm_["ssh_username"] = config.get_cloud_config_value(
            "ssh_username", vm_, __opts__
        )
    vm_["password"] = config.get_cloud_config_value("ssh_password", vm_, __opts__)
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)

    data = show_instance(vm_["name"], call="action")
    log.info("Created Cloud VM '%s'", vm_["name"])
    log.debug("'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data))

    ret.update(data)

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


@_deprecation_message
def destroy(name, call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    Destroy a VM.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -d myminion
        salt-cloud -a destroy myminion service_name=myservice
    """
    if kwargs is None:
        kwargs = {}

    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    compconn = get_conn(client_type="compute")

    node_data = show_instance(name, call="action")
    if node_data["storage_profile"]["os_disk"].get("managed_disk"):
        vhd = node_data["storage_profile"]["os_disk"]["managed_disk"]["id"]
    else:
        vhd = node_data["storage_profile"]["os_disk"]["vhd"]["uri"]

    ret = {name: {}}
    log.debug("Deleting VM")
    result = compconn.virtual_machines.delete(node_data["resource_group"], name)
    result.wait()

    if __opts__.get("update_cachedir", False) is True:
        __utils__["cloud.delete_minion_cachedir"](
            name, _get_active_provider_name().split(":")[0], __opts__
        )

    cleanup_disks = config.get_cloud_config_value(
        "cleanup_disks",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default=False,
    )

    if cleanup_disks:
        cleanup_vhds = kwargs.get(
            "delete_vhd",
            config.get_cloud_config_value(
                "cleanup_vhds",
                get_configured_provider(),
                __opts__,
                search_global=False,
                default=False,
            ),
        )

        if cleanup_vhds:
            log.debug("Deleting vhd")

            comps = vhd.split("/")
            container = comps[-2]
            blob = comps[-1]

            ret[name]["delete_disk"] = {
                "delete_disks": cleanup_disks,
                "delete_vhd": cleanup_vhds,
                "container": container,
                "blob": blob,
            }

            if vhd.startswith("http"):
                ret[name]["data"] = delete_blob(
                    kwargs={"container": container, "blob": blob}, call="function"
                )
            else:
                ret[name]["data"] = delete_managed_disk(
                    kwargs={
                        "resource_group": node_data["resource_group"],
                        "container": container,
                        "blob": blob,
                    },
                    call="function",
                )

        cleanup_data_disks = kwargs.get(
            "delete_data_disks",
            config.get_cloud_config_value(
                "cleanup_data_disks",
                get_configured_provider(),
                __opts__,
                search_global=False,
                default=False,
            ),
        )

        if cleanup_data_disks:
            log.debug("Deleting data_disks")
            ret[name]["data_disks"] = {}

            for disk in node_data["storage_profile"]["data_disks"]:
                datavhd = disk.get("managed_disk", {}).get("id") or disk.get(
                    "vhd", {}
                ).get("uri")
                comps = datavhd.split("/")
                container = comps[-2]
                blob = comps[-1]

                ret[name]["data_disks"][disk["name"]] = {
                    "delete_disks": cleanup_disks,
                    "delete_vhd": cleanup_vhds,
                    "container": container,
                    "blob": blob,
                }

                if datavhd.startswith("http"):
                    ret[name]["data"] = delete_blob(
                        kwargs={"container": container, "blob": blob}, call="function"
                    )
                else:
                    ret[name]["data"] = delete_managed_disk(
                        kwargs={
                            "resource_group": node_data["resource_group"],
                            "container": container,
                            "blob": blob,
                        },
                        call="function",
                    )

    cleanup_interfaces = config.get_cloud_config_value(
        "cleanup_interfaces",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default=False,
    )

    if cleanup_interfaces:
        ret[name]["cleanup_network"] = {
            "cleanup_interfaces": cleanup_interfaces,
            "resource_group": node_data["resource_group"],
            "data": [],
        }

        ifaces = node_data["network_profile"]["network_interfaces"]
        for iface in ifaces:
            resource_group = iface["id"].split("/")[4]
            ret[name]["cleanup_network"]["data"].append(
                delete_interface(
                    kwargs={
                        "resource_group": resource_group,
                        "iface_name": iface["name"],
                    },
                    call="function",
                )
            )

    return ret


@_deprecation_message
def list_storage_accounts(call=None):
    """
    List storage accounts within the subscription.
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_storage_accounts function must be called with -f or --function"
        )

    storconn = get_conn(client_type="storage")

    ret = {}
    try:
        accounts_query = storconn.storage_accounts.list()
        accounts = salt.utils.azurearm.paged_object_to_list(accounts_query)
        for account in accounts:
            ret[account["name"]] = account
    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error("storage", exc.message)
        ret = {"Error": exc.message}

    return ret


def _get_cloud_environment():
    """
    Get the cloud environment object.
    """
    cloud_environment = config.get_cloud_config_value(
        "cloud_environment", get_configured_provider(), __opts__, search_global=False
    )
    try:
        cloud_env_module = importlib.import_module("msrestazure.azure_cloud")
        cloud_env = getattr(cloud_env_module, cloud_environment or "AZURE_PUBLIC_CLOUD")
    except (AttributeError, ImportError):
        raise SaltCloudSystemExit(
            "The azure {} cloud environment is not available.".format(cloud_environment)
        )

    return cloud_env


def _get_block_blob_service(kwargs=None):
    """
    Get the block blob storage service.
    """
    resource_group = kwargs.get("resource_group") or config.get_cloud_config_value(
        "resource_group", get_configured_provider(), __opts__, search_global=False
    )
    sas_token = kwargs.get("sas_token") or config.get_cloud_config_value(
        "sas_token", get_configured_provider(), __opts__, search_global=False
    )
    storage_account = kwargs.get("storage_account") or config.get_cloud_config_value(
        "storage_account", get_configured_provider(), __opts__, search_global=False
    )
    storage_key = kwargs.get("storage_key") or config.get_cloud_config_value(
        "storage_key", get_configured_provider(), __opts__, search_global=False
    )

    if not resource_group:
        raise SaltCloudSystemExit("A resource group must be specified")

    if not storage_account:
        raise SaltCloudSystemExit("A storage account must be specified")

    if not storage_key:
        storconn = get_conn(client_type="storage")
        storage_keys = storconn.storage_accounts.list_keys(
            resource_group, storage_account
        )
        storage_keys = {v.key_name: v.value for v in storage_keys.keys}
        storage_key = next(iter(storage_keys.values()))

    cloud_env = _get_cloud_environment()

    endpoint_suffix = cloud_env.suffixes.storage_endpoint

    return BlockBlobService(
        storage_account,
        storage_key,
        sas_token=sas_token,
        endpoint_suffix=endpoint_suffix,
    )


@_deprecation_message
def list_blobs(call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    List blobs.
    """
    if kwargs is None:
        kwargs = {}

    if "container" not in kwargs:
        raise SaltCloudSystemExit("A container must be specified")

    storageservice = _get_block_blob_service(kwargs)

    ret = {}
    try:
        for blob in storageservice.list_blobs(kwargs["container"]).items:
            ret[blob.name] = {
                "blob_type": blob.properties.blob_type,
                "last_modified": blob.properties.last_modified.isoformat(),
                "server_encrypted": blob.properties.server_encrypted,
            }
    except Exception as exc:  # pylint: disable=broad-except
        log.warning(str(exc))

    return ret


@_deprecation_message
def delete_blob(call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    Delete a blob from a container.
    """
    if kwargs is None:
        kwargs = {}

    if "container" not in kwargs:
        raise SaltCloudSystemExit("A container must be specified")

    if "blob" not in kwargs:
        raise SaltCloudSystemExit("A blob must be specified")

    storageservice = _get_block_blob_service(kwargs)

    storageservice.delete_blob(kwargs["container"], kwargs["blob"])
    return True


@_deprecation_message
def delete_managed_disk(call=None, kwargs=None):  # pylint: disable=unused-argument
    """
    Delete a managed disk from a resource group.
    """

    compconn = get_conn(client_type="compute")

    try:
        compconn.disks.delete(kwargs["resource_group"], kwargs["blob"])
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error deleting managed disk %s - %s",
            kwargs.get("blob"),
            str(exc),
        )
        return False

    return True


@_deprecation_message
def list_virtual_networks(call=None, kwargs=None):
    """
    List virtual networks.
    """
    if kwargs is None:
        kwargs = {}

    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with -f or --function"
        )

    netconn = get_conn(client_type="network")
    resource_groups = list_resource_groups()

    ret = {}
    for group in resource_groups:
        try:
            networks = netconn.virtual_networks.list(resource_group_name=group)
        except CloudError:
            networks = {}
        for network_obj in networks:
            network = network_obj.as_dict()
            ret[network["name"]] = network
            ret[network["name"]]["subnets"] = list_subnets(
                kwargs={"resource_group": group, "network": network["name"]}
            )

    return ret


@_deprecation_message
def list_subnets(call=None, kwargs=None):
    """
    List subnets in a virtual network.
    """
    if kwargs is None:
        kwargs = {}

    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with -f or --function"
        )

    netconn = get_conn(client_type="network")

    resource_group = kwargs.get("resource_group") or config.get_cloud_config_value(
        "resource_group", get_configured_provider(), __opts__, search_global=False
    )

    if not resource_group and "group" in kwargs and "resource_group" not in kwargs:
        resource_group = kwargs["group"]

    if not resource_group:
        raise SaltCloudSystemExit("A resource group must be specified")

    if kwargs.get("network") is None:
        kwargs["network"] = config.get_cloud_config_value(
            "network", get_configured_provider(), __opts__, search_global=False
        )

    if "network" not in kwargs or kwargs["network"] is None:
        raise SaltCloudSystemExit('A "network" must be specified')

    ret = {}
    subnets = netconn.subnets.list(resource_group, kwargs["network"])
    for subnet in subnets:
        ret[subnet.name] = subnet.as_dict()
        ret[subnet.name]["ip_configurations"] = {}
        for ip_ in subnet.ip_configurations:
            comps = ip_.id.split("/")
            name = comps[-1]
            ret[subnet.name]["ip_configurations"][name] = ip_.as_dict()
            ret[subnet.name]["ip_configurations"][name]["subnet"] = subnet.name
        ret[subnet.name]["resource_group"] = resource_group
    return ret


@_deprecation_message
def create_or_update_vmextension(
    call=None, kwargs=None
):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2019.2.0

    Create or update a VM extension object "inside" of a VM object.

    required kwargs:
      .. code-block:: yaml

        extension_name: myvmextension
        virtual_machine_name: myvm
        settings: {"commandToExecute": "hostname"}

    optional kwargs:
      .. code-block:: yaml

        resource_group: < inferred from cloud configs >
        location: < inferred from cloud configs >
        publisher: < default: Microsoft.Azure.Extensions >
        virtual_machine_extension_type: < default: CustomScript >
        type_handler_version: < default: 2.0 >
        auto_upgrade_minor_version: < default: True >
        protected_settings: < default: None >
    """
    if kwargs is None:
        kwargs = {}

    if "extension_name" not in kwargs:
        raise SaltCloudSystemExit("An extension name must be specified")

    if "virtual_machine_name" not in kwargs:
        raise SaltCloudSystemExit("A virtual machine name must be specified")

    compconn = get_conn(client_type="compute")

    # pylint: disable=invalid-name
    VirtualMachineExtension = getattr(compute_models, "VirtualMachineExtension")

    resource_group = kwargs.get("resource_group") or config.get_cloud_config_value(
        "resource_group", get_configured_provider(), __opts__, search_global=False
    )

    if not resource_group:
        raise SaltCloudSystemExit("A resource group must be specified")

    location = kwargs.get("location") or get_location()

    if not location:
        raise SaltCloudSystemExit("A location must be specified")

    publisher = kwargs.get("publisher", "Microsoft.Azure.Extensions")
    virtual_machine_extension_type = kwargs.get(
        "virtual_machine_extension_type", "CustomScript"
    )
    type_handler_version = kwargs.get("type_handler_version", "2.0")
    auto_upgrade_minor_version = kwargs.get("auto_upgrade_minor_version", True)
    settings = kwargs.get("settings", {})
    protected_settings = kwargs.get("protected_settings")

    if not isinstance(settings, dict):
        raise SaltCloudSystemExit("VM extension settings are not valid")
    elif "commandToExecute" not in settings and "script" not in settings:
        raise SaltCloudSystemExit(
            "VM extension settings are not valid. Either commandToExecute or script"
            " must be specified."
        )

    log.info("Creating VM extension %s", kwargs["extension_name"])

    ret = {}
    try:
        params = VirtualMachineExtension(
            location=location,
            publisher=publisher,
            virtual_machine_extension_type=virtual_machine_extension_type,
            type_handler_version=type_handler_version,
            auto_upgrade_minor_version=auto_upgrade_minor_version,
            settings=settings,
            protected_settings=protected_settings,
        )
        poller = compconn.virtual_machine_extensions.create_or_update(
            resource_group,
            kwargs["virtual_machine_name"],
            kwargs["extension_name"],
            params,
        )
        ret = poller.result()
        ret = ret.as_dict()

    except CloudError as exc:
        salt.utils.azurearm.log_cloud_error(
            "compute",
            "Error attempting to create the VM extension: {}".format(exc.message),
        )
        ret = {"error": exc.message}

    return ret


@_deprecation_message
def stop(name, call=None):
    """
    .. versionadded:: 2019.2.0

    Stop (deallocate) a VM

    CLI Examples:

    .. code-block:: bash

         salt-cloud -a stop myminion
    """
    if call == "function":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    compconn = get_conn(client_type="compute")

    resource_group = config.get_cloud_config_value(
        "resource_group", get_configured_provider(), __opts__, search_global=False
    )

    ret = {}
    if not resource_group:
        groups = list_resource_groups()
        for group in groups:
            try:
                instance = compconn.virtual_machines.deallocate(
                    vm_name=name, resource_group_name=group
                )
                instance.wait()
                vm_result = instance.result()
                ret = vm_result.as_dict()
                break
            except CloudError as exc:
                if "was not found" in exc.message:
                    continue
                else:
                    ret = {"error": exc.message}
        if not ret:
            salt.utils.azurearm.log_cloud_error(
                "compute", "Unable to find virtual machine with name: {}".format(name)
            )
            ret = {"error": "Unable to find virtual machine with name: {}".format(name)}
    else:
        try:
            instance = compconn.virtual_machines.deallocate(
                vm_name=name, resource_group_name=resource_group
            )
            instance.wait()
            vm_result = instance.result()
            ret = vm_result.as_dict()
        except CloudError as exc:
            salt.utils.azurearm.log_cloud_error(
                "compute", "Error attempting to stop {}: {}".format(name, exc.message)
            )
            ret = {"error": exc.message}

    return ret


@_deprecation_message
def start(name, call=None):
    """
    .. versionadded:: 2019.2.0

    Start a VM

    CLI Examples:

    .. code-block:: bash

         salt-cloud -a start myminion
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The start action must be called with -a or --action."
        )

    compconn = get_conn(client_type="compute")

    resource_group = config.get_cloud_config_value(
        "resource_group", get_configured_provider(), __opts__, search_global=False
    )

    ret = {}
    if not resource_group:
        groups = list_resource_groups()
        for group in groups:
            try:
                instance = compconn.virtual_machines.start(
                    vm_name=name, resource_group_name=group
                )
                instance.wait()
                vm_result = instance.result()
                ret = vm_result.as_dict()
                break
            except CloudError as exc:
                if "was not found" in exc.message:
                    continue
                else:
                    ret = {"error": exc.message}
        if not ret:
            salt.utils.azurearm.log_cloud_error(
                "compute", "Unable to find virtual machine with name: {}".format(name)
            )
            ret = {"error": "Unable to find virtual machine with name: {}".format(name)}
    else:
        try:
            instance = compconn.virtual_machines.start(
                vm_name=name, resource_group_name=resource_group
            )
            instance.wait()
            vm_result = instance.result()
            ret = vm_result.as_dict()
        except CloudError as exc:
            salt.utils.azurearm.log_cloud_error(
                "compute",
                "Error attempting to start {}: {}".format(name, exc.message),
            )
            ret = {"error": exc.message}

    return ret



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import pytest

import salt.states.event as event
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {event: {}}


def test_send():
    """
    Test to send an event to the Salt Master
    """
    with patch.dict(event.__opts__, {"test": True}):
        assert event.send("salt") == {
            "changes": {"data": None, "tag": "salt"},
            "comment": "Event would have been fired",
            "name": "salt",
            "result": None,
        }

    with patch.dict(event.__opts__, {"test": False}):
        mock = MagicMock(return_value=True)
        with patch.dict(event.__salt__, {"event.send": mock}):
            assert event.send("salt") == {
                "changes": {"data": None, "tag": "salt"},
                "comment": "Event fired",
                "name": "salt",
                "result": True,
            }


def test_wait():
    """
    Test to fire an event on the Salt master
    """
    assert event.wait("salt") == {
        "changes": {},
        "comment": "",
        "name": "salt",
        "result": True,
    }

    assert event.wait("salt", sfun="some_function") == {
        "changes": {},
        "comment": "",
        "name": "salt",
        "result": True,
    }

    assert event.wait("salt", data={"random": "data"}) == {
        "changes": {},
        "comment": "",
        "name": "salt",
        "result": True,
    }



"""
Module to interact with Junos devices.

:maturity: new
:dependencies: junos-eznc, jxmlease

.. note::

    Those who wish to use junos-eznc (PyEZ) version >= 2.1.0, must
    use the latest salt code from github until the next release.

Refer to :mod:`junos <salt.proxy.junos>` for information on connecting to junos proxy.

"""

import copy
import json
import logging
import os
import re
from functools import wraps

import yaml

import salt.utils.args
import salt.utils.files
import salt.utils.json
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree


# Juniper interface libraries
# https://github.com/Juniper/py-junos-eznc
try:
    # pylint: disable=W0611
    import jnpr.junos.cfg
    import jnpr.junos.op as tables_dir
    import jnpr.junos.utils
    import jxmlease
    import yamlordereddictloader
    from jnpr.junos import Device
    from jnpr.junos.exception import (
        ConnectClosedError,
        LockError,
        RpcTimeoutError,
        UnlockError,
    )
    from jnpr.junos.factory.cfgtable import CfgTable
    from jnpr.junos.factory.factory_loader import FactoryLoader
    from jnpr.junos.factory.optable import OpTable
    from jnpr.junos.utils.config import Config
    from jnpr.junos.utils.scp import SCP
    from jnpr.junos.utils.sw import SW

    # pylint: enable=W0611
    HAS_JUNOS = True
except ImportError:
    HAS_JUNOS = False

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "junos"

__proxyenabled__ = ["junos"]


def __virtual__():
    """
    We need the Junos adapter libraries for this
    module to work.  We also need a proxymodule entry in __opts__
    in the opts dictionary
    """
    if HAS_JUNOS and "proxy" in __opts__:
        return __virtualname__
    else:
        return (
            False,
            "The junos or dependent module could not be loaded: "
            "junos-eznc or jxmlease or yamlordereddictloader or "
            "proxy could not be loaded.",
        )


class HandleFileCopy:
    """
    To figure out proper path either from proxy local file system
    or proxy cache or on master. If required, then only copy from
    master to proxy

    """

    def __init__(self, path, **kwargs):
        self._file_path = path
        self._cached_folder = None
        self._cached_file = None
        self._kwargs = kwargs

    def __enter__(self):
        if self._file_path.startswith("salt://"):
            # check if file exists in cache
            local_cache_path = __salt__["cp.is_cached"](self._file_path)
            if local_cache_path:
                master_hash = __salt__["cp.hash_file"](self._file_path)
                proxy_hash = __salt__["file.get_hash"](local_cache_path)
                # check if hash is same, else copy newly
                if master_hash.get("hsum") == proxy_hash:
                    self._cached_file = salt.utils.files.mkstemp()
                    # local copy is a template, hence need to render
                    with salt.utils.files.fopen(self._cached_file, "w") as fp:
                        template_string = __salt__["slsutil.renderer"](
                            path=local_cache_path,
                            default_renderer="jinja",
                            **self._kwargs,
                        )
                        fp.write(template_string)
                    return self._cached_file

            # continue for else part
            self._cached_file = salt.utils.files.mkstemp()
            __salt__["cp.get_template"](
                self._file_path, self._cached_file, **self._kwargs
            )
            if self._cached_file != "":
                return self._cached_file
        else:
            # check for local location of file
            if __salt__["file.file_exists"](self._file_path):
                self._cached_file = salt.utils.files.mkstemp()
                with salt.utils.files.fopen(self._cached_file, "w") as fp:
                    template_string = __salt__["slsutil.renderer"](
                        path=self._file_path, default_renderer="jinja", **self._kwargs
                    )
                    fp.write(template_string)
                return self._cached_file

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self._cached_file is not None:
            salt.utils.files.safe_rm(self._cached_file)
            log.debug("Deleted cached file: %s", self._cached_file)
        if self._cached_folder is not None:
            __salt__["file.rmdir"](self._cached_folder)
            log.debug("Deleted cached folder: %s", self._cached_folder)


def _timeout_decorator(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        if "dev_timeout" in kwargs or "timeout" in kwargs:
            ldev_timeout = max(kwargs.pop("dev_timeout", 0), kwargs.pop("timeout", 0))
            conn = __proxy__["junos.conn"]()
            restore_timeout = conn.timeout
            conn.timeout = ldev_timeout
            try:
                result = function(*args, **kwargs)
                conn.timeout = restore_timeout
                return result
            except Exception:  # pylint: disable=broad-except
                conn.timeout = restore_timeout
                raise
        else:
            return function(*args, **kwargs)

    return wrapper


def _timeout_decorator_cleankwargs(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        if "dev_timeout" in kwargs or "timeout" in kwargs:
            ldev_timeout = max(kwargs.pop("dev_timeout", 0), kwargs.pop("timeout", 0))
            conn = __proxy__["junos.conn"]()
            restore_timeout = conn.timeout
            conn.timeout = ldev_timeout
            try:
                restore_kwargs = False
                del_list = []
                op = {}
                op.update(kwargs)
                for keychk in kwargs:
                    if keychk.startswith("__pub"):
                        del_list.append(keychk)
                if del_list:
                    restore_kwargs = True
                    for delkey in del_list:
                        kwargs.pop(delkey)

                result = function(*args, **kwargs)
                if restore_kwargs:
                    kwargs.update(op)

                conn.timeout = restore_timeout
                return result
            except Exception:  # pylint: disable=broad-except
                conn.timeout = restore_timeout
                raise
        else:
            restore_kwargs = False
            del_list = []
            op = {}
            op.update(kwargs)
            for keychk in kwargs:
                if keychk.startswith("__pub"):
                    del_list.append(keychk)
            if del_list:
                restore_kwargs = True
                for delkey in del_list:
                    kwargs.pop(delkey)

            ret = function(*args, **kwargs)
            if restore_kwargs:
                kwargs.update(op)

            return ret

    return wrapper


def _restart_connection():
    minion_id = __opts__.get("proxyid", "") or __opts__.get("id", "")
    log.info(
        "Junos exception occurred %s (junos proxy) is down. Restarting.", minion_id
    )
    __salt__["event.fire_master"](
        {}, "junos/proxy/{}/stop".format(__opts__["proxy"]["host"])
    )
    __proxy__["junos.shutdown"](__opts__)  # safely close connection
    __proxy__["junos.init"](__opts__)  # reopen connection
    log.debug("Junos exception occurred, restarted %s (junos proxy)!", minion_id)


@_timeout_decorator_cleankwargs
def facts_refresh():
    """
    Reload the facts dictionary from the device. Usually only needed if,
    the device configuration is changed by some other actor.
    This function will also refresh the facts stored in the salt grains.

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.facts_refresh
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    try:
        conn.facts_refresh()
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Execution failed due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()
        return ret

    ret["facts"] = __proxy__["junos.get_serialized_facts"]()

    try:
        __salt__["saltutil.sync_grains"]()
    except Exception as exception:  # pylint: disable=broad-except
        log.error('Grains could not be updated due to "%s"', exception)

    return ret


def facts():
    """
    Displays the facts gathered during the connection.
    These facts are also stored in Salt grains.

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.facts
    """
    ret = {}
    try:
        ret["facts"] = __proxy__["junos.get_serialized_facts"]()
        ret["out"] = True
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not display facts due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()

    return ret


@_timeout_decorator
def rpc(cmd=None, dest=None, **kwargs):
    """
    This function executes the RPC provided as arguments on the junos device.
    The returned data can be stored in a file.

    cmd
        The RPC to be executed

    dest
        Destination file where the RPC output is stored. Note that the file
        will be stored on the proxy minion. To push the files to the master use
        :py:func:`cp.push <salt.modules.cp.push>`.

    format : xml
        The format in which the RPC reply is received from the device

    dev_timeout : 30
        The NETCONF RPC timeout (in seconds)

    filter
        Used with the ``get-config`` RPC to get specific configuration

    terse : False
        Amount of information you want

    interface_name
      Name of the interface to query

    CLI Example:

    .. code-block:: bash

        salt 'device' junos.rpc get_config dest=/var/log/config.txt format=text filter='<configuration><system/></configuration>'
        salt 'device' junos.rpc get-interface-information dest=/home/user/interface.xml interface_name='lo0' terse=True
        salt 'device' junos.rpc get-chassis-inventory
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True

    op = dict()
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    elif "__pub_schedule" in kwargs:
        for key, value in kwargs.items():
            if not key.startswith("__pub_"):
                op[key] = value
    else:
        op.update(kwargs)

    if cmd is None:
        ret["message"] = "Please provide the rpc to execute."
        ret["out"] = False
        return ret

    format_ = op.pop("format", "xml")
    # dest becomes part of op via __pub_arg if not None
    # rpc commands objecting to dest as part of op
    op.pop("dest", dest)

    if cmd in ["get-config", "get_config"]:
        filter_reply = None
        if "filter" in op:
            try:
                filter_reply = etree.XML(op["filter"])
            except etree.XMLSyntaxError as ex:
                ret["message"] = "Invalid filter: {}".format(str(ex))
                ret["out"] = False
                return ret

            del op["filter"]

        op.update({"format": format_})
        try:
            reply = getattr(conn.rpc, cmd.replace("-", "_"))(filter_reply, options=op)
        except Exception as exception:  # pylint: disable=broad-except
            ret["message"] = 'RPC execution failed due to "{}"'.format(exception)
            ret["out"] = False
            _restart_connection()
            return ret
    else:
        if "filter" in op:
            log.warning('Filter ignored as it is only used with "get-config" rpc')

        if "dest" in op:
            log.warning("dest in op, rpc may reject this for cmd '%s'", cmd)

        try:
            reply = getattr(conn.rpc, cmd.replace("-", "_"))({"format": format_}, **op)
        except Exception as exception:  # pylint: disable=broad-except
            ret["message"] = 'RPC execution failed due to "{}"'.format(exception)
            ret["out"] = False
            _restart_connection()
            return ret

    if format_ == "text":
        ret["rpc_reply"] = reply.text
    elif format_ == "json":
        ret["rpc_reply"] = reply
    else:
        ret["rpc_reply"] = jxmlease.parse(etree.tostring(reply))

    if dest:
        if format_ == "text":
            write_response = reply.text
        elif format_ == "json":
            write_response = salt.utils.json.dumps(reply, indent=1)
        else:
            write_response = etree.tostring(reply)
        with salt.utils.files.fopen(dest, "w") as fp:
            fp.write(salt.utils.stringutils.to_str(write_response))

    return ret


@_timeout_decorator
def set_hostname(hostname=None, **kwargs):
    """
    Set the device's hostname

    hostname
        The name to be set

    comment
        Provide a comment to the commit

    dev_timeout : 30
        The NETCONF RPC timeout (in seconds)

    confirm
      Provide time in minutes for commit confirmation. If this option is
      specified, the commit will be rolled back in the specified amount of time
      unless the commit is confirmed.

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.set_hostname salt-device
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    if hostname is None:
        ret["message"] = "Please provide the hostname."
        ret["out"] = False
        return ret

    op = dict()
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    # Added to recent versions of JunOs
    # Use text format instead
    set_string = "set system host-name {}".format(hostname)
    try:
        conn.cu.load(set_string, format="set")
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not load configuration due to error "{}"'.format(
            exception
        )
        ret["out"] = False
        _restart_connection()
        return ret

    try:
        commit_ok = conn.cu.commit_check()
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not commit check due to error "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()
        return ret

    if commit_ok:
        try:
            conn.cu.commit(**op)
            ret["message"] = "Successfully changed hostname."
            ret["out"] = True
        except Exception as exception:  # pylint: disable=broad-except
            ret["out"] = False
            ret[
                "message"
            ] = 'Successfully loaded host-name but commit failed with "{}"'.format(
                exception
            )
            _restart_connection()
            return ret
    else:
        ret["out"] = False
        ret["message"] = "Successfully loaded host-name but pre-commit check failed."
        try:
            conn.cu.rollback()
        except Exception as exception:  # pylint: disable=broad-except
            ret["out"] = False
            ret[
                "message"
            ] = 'Successfully loaded host-name but rollback before exit failed "{}"'.format(
                exception
            )
            _restart_connection()

    return ret


@_timeout_decorator
def commit(**kwargs):
    """
    To commit the changes loaded in the candidate configuration.

    dev_timeout : 30
        The NETCONF RPC timeout (in seconds)

    comment
      Provide a comment for the commit

    confirm
      Provide time in minutes for commit confirmation. If this option is
      specified, the commit will be rolled back in the specified amount of time
      unless the commit is confirmed.

    sync : False
      When ``True``, on dual control plane systems, requests that the candidate
      configuration on one control plane be copied to the other control plane,
      checked for correct syntax, and committed on both Routing Engines.

    force_sync : False
      When ``True``, on dual control plane systems, force the candidate
      configuration on one control plane to be copied to the other control
      plane.

    full
      When ``True``, requires all the daemons to check and evaluate the new
      configuration.

    detail
      When ``True``, return commit detail

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.commit comment='Commiting via saltstack' detail=True
        salt 'device_name' junos.commit dev_timeout=60 confirm=10
        salt 'device_name' junos.commit sync=True dev_timeout=90
    """

    conn = __proxy__["junos.conn"]()
    ret = {}
    op = dict()
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    op["detail"] = op.get("detail", False)

    try:
        commit_ok = conn.cu.commit_check()
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not perform commit check due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()
        return ret

    if commit_ok:
        try:
            commit = conn.cu.commit(**op)
            ret["out"] = True
            if commit:
                if op["detail"]:
                    ret["message"] = jxmlease.parse(etree.tostring(commit))
                else:
                    ret["message"] = "Commit Successful."
            else:
                ret["message"] = "Commit failed."
                ret["out"] = False
        except Exception as exception:  # pylint: disable=broad-except
            ret["out"] = False
            ret[
                "message"
            ] = 'Commit check succeeded but actual commit failed with "{}"'.format(
                exception
            )
            _restart_connection()
    else:
        ret["out"] = False
        ret["message"] = "Pre-commit check failed."
        try:
            conn.cu.rollback()
        except Exception as exception:  # pylint: disable=broad-except
            ret["out"] = False
            ret[
                "message"
            ] = 'Pre-commit check failed, and exception during rollback "{}"'.format(
                exception
            )
            _restart_connection()

    return ret


@_timeout_decorator
def rollback(**kwargs):
    """
    Roll back the last committed configuration changes and commit

    id : 0
        The rollback ID value (0-49)

    d_id : 0
        The rollback ID value (0-49)

    dev_timeout : 30
        The NETCONF RPC timeout (in seconds)

    comment
      Provide a comment for the commit

    confirm
      Provide time in minutes for commit confirmation. If this option is
      specified, the commit will be rolled back in the specified amount of time
      unless the commit is confirmed.

    diffs_file
      Path to the file where the diff (difference in old configuration and the
      committed configuration) will be stored. Note that the file will be
      stored on the proxy minion. To push the files to the master use
      :py:func:`cp.push <salt.modules.cp.push>`.

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.rollback 10

    NOTE: Because of historical reasons and the internals of the Salt state
    compiler, there are three possible sources of the rollback ID--the
    positional argument, and the `id` and `d_id` kwargs.  The precedence of
    the arguments are `id` (positional), `id` (kwarg), `d_id` (kwarg).  In
    other words, if all three are passed, only the positional argument
    will be used.  A warning is logged if more than one is passed.
    """
    ids_passed = 0
    id_ = 0
    if "d_id" in kwargs:
        id_ = kwargs.pop("d_id")
        ids_passed = ids_passed + 1
    if "id" in kwargs:
        id_ = kwargs.pop("id", 0)
        ids_passed = ids_passed + 1

    if ids_passed > 1:
        log.warning(
            "junos.rollback called with more than one possible ID. "
            "Use only one of the positional argument, `id`, or `d_id` kwargs"
        )

    ret = {}
    conn = __proxy__["junos.conn"]()

    op = dict()
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    try:
        ret["out"] = conn.cu.rollback(id_)
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Rollback failed due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()
        return ret

    if ret["out"]:
        ret["message"] = "Rollback successful"
    else:
        ret["message"] = "Rollback failed"
        return ret

    if "diffs_file" in op and op["diffs_file"] is not None:
        diff = conn.cu.diff()
        if diff is not None:
            with salt.utils.files.fopen(op["diffs_file"], "w") as fp:
                fp.write(salt.utils.stringutils.to_str(diff))
        else:
            log.info(
                "No diff between current configuration and "
                "rollbacked configuration, so no diff file created"
            )

    try:
        commit_ok = conn.cu.commit_check()
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not commit check due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()
        return ret

    if commit_ok:
        try:
            conn.cu.commit(**op)
            ret["out"] = True
        except Exception as exception:  # pylint: disable=broad-except
            ret["out"] = False
            ret[
                "message"
            ] = 'Rollback successful but commit failed with error "{}"'.format(
                exception
            )
            _restart_connection()
            return ret
    else:
        ret["message"] = "Rollback successful but pre-commit check failed."
        ret["out"] = False

    return ret


@_timeout_decorator
def diff(**kwargs):
    """
    Returns the difference between the candidate and the current configuration

    id : 0
        The rollback ID value (0-49)

    d_id : 0
        The rollback ID value (0-49)

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.diff d_id=3

    NOTE: Because of historical reasons and the internals of the Salt state
    compiler, there are three possible sources of the rollback ID--the
    positional argument, and the `id` and `d_id` kwargs.  The precedence of
    the arguments are `id` (positional), `id` (kwarg), `d_id` (kwarg).  In
    other words, if all three are passed, only the positional argument
    will be used.  A warning is logged if more than one is passed.
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    ids_passed = 0
    id_ = 0
    if "d_id" in kwargs:
        id_ = kwargs.pop("d_id")
        ids_passed = ids_passed + 1
    if "id" in kwargs:
        id_ = kwargs.pop("id", 0)
        ids_passed = ids_passed + 1
    if ids_passed > 1:
        log.warning(
            "junos.rollback called with more than one possible ID. "
            "Use only one of the positional argument, `id`, or `d_id` kwargs"
        )

    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    try:
        ret["message"] = conn.cu.diff(rb_id=id_)
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not get diff with error "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()

    return ret


@_timeout_decorator
def ping(dest_ip=None, **kwargs):
    """
    Send a ping RPC to a device

    dest_ip
      The IP of the device to ping

    dev_timeout : 30
        The NETCONF RPC timeout (in seconds)

    rapid : False
        When ``True``, executes ping at 100pps instead of 1pps

    ttl
        Maximum number of IP routers (IP hops) allowed between source and
        destination

    routing_instance
      Name of the routing instance to use to send the ping

    interface
      Interface used to send traffic

    count : 5
      Number of packets to send

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.ping '8.8.8.8' count=5
        salt 'device_name' junos.ping '8.8.8.8' ttl=1 rapid=True
    """
    conn = __proxy__["junos.conn"]()
    ret = {}

    if dest_ip is None:
        ret["message"] = "Please specify the destination ip to ping."
        ret["out"] = False
        return ret

    op = {"host": dest_ip}
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    op["count"] = str(op.pop("count", 5))
    if "ttl" in op:
        op["ttl"] = str(op["ttl"])

    ret["out"] = True
    try:
        ret["message"] = jxmlease.parse(etree.tostring(conn.rpc.ping(**op)))
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Execution failed due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()

    return ret


@_timeout_decorator
def cli(command=None, **kwargs):
    """
    Executes the CLI commands and returns the output in specified format. \
    (default is text) The output can also be stored in a file.

    command (required)
        The command to execute on the Junos CLI

    format : text
        Format in which to get the CLI output (either ``text`` or ``xml``)

    dev_timeout : 30
        The NETCONF RPC timeout (in seconds)

    dest
        Destination file where the RPC output is stored. Note that the file
        will be stored on the proxy minion. To push the files to the master use
        :py:func:`cp.push <salt.modules.cp.push>`.

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.cli 'show system commit'
        salt 'device_name' junos.cli 'show system alarms' format=xml dest=/home/user/cli_output.txt
    """
    conn = __proxy__["junos.conn"]()

    format_ = kwargs.pop("format", "text")
    if not format_:
        format_ = "text"

    ret = {}
    if command is None:
        ret["message"] = "Please provide the CLI command to be executed."
        ret["out"] = False
        return ret

    op = dict()
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    try:
        result = conn.cli(command, format_, warning=False)
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Execution failed due to "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()
        return ret

    if format_ == "text":
        ret["message"] = result
    else:
        result = etree.tostring(result)
        ret["message"] = jxmlease.parse(result)

    if "dest" in op and op["dest"] is not None:
        try:
            with salt.utils.files.fopen(op["dest"], "w") as fp:
                fp.write(salt.utils.stringutils.to_str(result))
        except OSError:
            ret["message"] = 'Unable to open "{}" to write'.format(op["dest"])
            ret["out"] = False
            return ret

    ret["out"] = True
    return ret


@_timeout_decorator
def shutdown(**kwargs):
    """
    Shut down (power off) or reboot a device running Junos OS. This includes
    all Routing Engines in a Virtual Chassis or a dual Routing Engine system.

      .. note::
          One of ``shutdown`` or ``reboot`` must be set to ``True`` or no
          action will be taken.

    shutdown : False
      Set this to ``True`` if you want to shutdown the machine. This is a
      safety mechanism so that the user does not accidentally shutdown the
      junos device.

    reboot : False
      If ``True``, reboot instead of shutting down

    at
      Used when rebooting, to specify the date and time the reboot should take
      place. The value of this option must match the JunOS CLI reboot syntax.

    in_min
        Used when shutting down. Specify the delay (in minutes) before the
        device will be shut down.

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.shutdown reboot=True
        salt 'device_name' junos.shutdown shutdown=True in_min=10
        salt 'device_name' junos.shutdown shutdown=True
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    sw = SW(conn)

    op = {}
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)
    if "shutdown" not in op and "reboot" not in op:
        ret["message"] = "Provide either one of the arguments: shutdown or reboot."
        ret["out"] = False
        return ret

    try:
        if "reboot" in op and op["reboot"]:
            shut = sw.reboot
        elif "shutdown" in op and op["shutdown"]:
            shut = sw.poweroff
        else:
            ret["message"] = "Nothing to be done."
            ret["out"] = False
            return ret

        if "in_min" in op:
            shut(in_min=op["in_min"])
        elif "at" in op:
            shut(at=op["at"])
        else:
            shut()
        ret["message"] = "Successfully powered off/rebooted."
        ret["out"] = True
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not poweroff/reboot because "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()

    return ret


@_timeout_decorator
def install_config(path=None, **kwargs):
    """
    Installs the given configuration file into the candidate configuration.
    Commits the changes if the commit checks or throws an error.

    path (required)
        Path where the configuration/template file is present. If the file has
        a ``.conf`` extension, the content is treated as text format. If the
        file has a ``.xml`` extension, the content is treated as XML format. If
        the file has a ``.set`` extension, the content is treated as Junos OS
        ``set`` commands.

    mode : exclusive
        The mode in which the configuration is locked. Can be one of
        ``private``, ``dynamic``, ``batch``, ``exclusive``, ``ephemeral``

    dev_timeout : 30
        Set NETCONF RPC timeout. Can be used for commands which take a while to
        execute.

    overwrite : False
        Set to ``True`` if you want this file is to completely replace the
        configuration file. Sets action to override

        .. note:: This option cannot be used if **format** is "set".

    replace : False
        Specify whether the configuration file uses ``replace:`` statements. If
        ``True``, only those statements under the ``replace`` tag will be
        changed.

    merge : False
        If set to ``True`` will set the load-config action to merge.
        the default load-config action is 'replace' for xml/json/text config

    format
        Determines the format of the contents

    update : False
        Compare a complete loaded configuration against the candidate
        configuration. For each hierarchy level or configuration object that is
        different in the two configurations, the version in the loaded
        configuration replaces the version in the candidate configuration. When
        the configuration is later committed, only system processes that are
        affected by the changed configuration elements parse the new
        configuration. This action is supported from PyEZ 2.1.

    comment
      Provide a comment for the commit

    confirm
      Provide time in minutes for commit confirmation. If this option is
      specified, the commit will be rolled back in the specified amount of time
      unless the commit is confirmed.

    diffs_file
      Path to the file where the diff (difference in old configuration and the
      committed configuration) will be stored. Note that the file will be
      stored on the proxy minion. To push the files to the master use:

        py:func:`cp.push <salt.modules.cp.push>`.

    template_vars
      Variables to be passed into the template processing engine in addition to
      those present in pillar, the minion configuration, grains, etc.  You may
      reference these variables in your template like so:

      .. code-block:: jinja

          {{ template_vars["var_name"] }}

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.install_config 'salt://production/network/routers/config.set'
        salt 'device_name' junos.install_config 'salt://templates/replace_config.conf' replace=True comment='Committed via SaltStack'
        salt 'device_name' junos.install_config 'salt://my_new_configuration.conf' dev_timeout=300 diffs_file='/salt/confs/old_config.conf' overwrite=True
        salt 'device_name' junos.install_config 'salt://syslog_template.conf' template_vars='{"syslog_host": "10.180.222.7"}'
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True

    if path is None:
        ret[
            "message"
        ] = "Please provide the salt path where the configuration is present"
        ret["out"] = False
        return ret

    op = {}
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    test = op.pop("test", False)

    kwargs = {}
    if "template_vars" in op:
        kwargs.update({"template_vars": op["template_vars"]})

    with HandleFileCopy(path, **kwargs) as template_cached_path:
        if template_cached_path is None:
            ret["message"] = "Invalid file path."
            ret["out"] = False
            return ret

        if os.path.getsize(template_cached_path) == 0:
            ret["message"] = "Template failed to render"
            ret["out"] = False
            return ret

        write_diff = ""
        if "diffs_file" in op and op["diffs_file"] is not None:
            write_diff = op["diffs_file"]
            del op["diffs_file"]

        op["path"] = template_cached_path

        if "format" not in op:
            if path.endswith("set"):
                template_format = "set"
            elif path.endswith("xml"):
                template_format = "xml"
            elif path.endswith("json"):
                template_format = "json"
            else:
                template_format = "text"

            op["format"] = template_format

        if "replace" in op and op["replace"]:
            op["merge"] = False
            del op["replace"]
        elif "overwrite" in op and op["overwrite"]:
            op["overwrite"] = True
        elif "overwrite" in op and not op["overwrite"]:
            op["merge"] = True
            del op["overwrite"]

        db_mode = op.pop("mode", "exclusive")
        if write_diff and db_mode in ["dynamic", "ephemeral"]:
            ret[
                "message"
            ] = "Write diff is not supported with dynamic/ephemeral configuration mode"
            ret["out"] = False
            return ret

        config_params = {}
        if "ephemeral_instance" in op:
            config_params["ephemeral_instance"] = op.pop("ephemeral_instance")
        try:
            with Config(conn, mode=db_mode, **config_params) as cu:
                try:
                    cu.load(**op)
                except Exception as exception:  # pylint: disable=broad-except
                    ret[
                        "message"
                    ] = 'Could not load configuration due to : "{}"'.format(exception)
                    ret["format"] = op["format"]
                    ret["out"] = False
                    _restart_connection()
                    return ret

                config_diff = None
                if db_mode in ["dynamic", "ephemeral"]:
                    log.warning("diff is not supported for dynamic and ephemeral")
                else:
                    config_diff = cu.diff()
                    if config_diff is None:
                        ret["message"] = "Configuration already applied!"
                        ret["out"] = True
                        return ret

                commit_params = {}
                if "confirm" in op:
                    commit_params["confirm"] = op["confirm"]
                if "comment" in op:
                    commit_params["comment"] = op["comment"]

                # Assume commit_check succeeds and initialize variable check
                check = True
                if db_mode in ["dynamic", "ephemeral"]:
                    log.warning("commit check not supported for dynamic and ephemeral")
                else:
                    try:
                        check = cu.commit_check()
                    except Exception as exception:  # pylint: disable=broad-except
                        ret[
                            "message"
                        ] = 'Commit check threw the following exception: "{}"'.format(
                            exception
                        )
                        ret["out"] = False
                        _restart_connection()
                        return ret

                if check and not test:
                    try:
                        cu.commit(**commit_params)
                        ret["message"] = "Successfully loaded and committed!"
                    except Exception as exception:  # pylint: disable=broad-except
                        ret[
                            "message"
                        ] = 'Commit check successful but commit failed with "{}"'.format(
                            exception
                        )
                        ret["out"] = False
                        _restart_connection()
                        return ret

                elif not check:
                    try:
                        cu.rollback()
                        ret["message"] = (
                            "Loaded configuration but commit check failed, hence"
                            " rolling back configuration."
                        )
                    except Exception as exception:  # pylint: disable=broad-except
                        ret["message"] = (
                            "Loaded configuration but commit check failed, and"
                            ' exception occurred during rolling back configuration "{}"'.format(
                                exception
                            )
                        )
                        _restart_connection()

                    ret["out"] = False
                else:
                    try:
                        cu.rollback()
                        ret["message"] = (
                            "Commit check passed, but skipping commit for dry-run and"
                            " rolling back configuration."
                        )
                        ret["out"] = True
                    except Exception as exception:  # pylint: disable=broad-except
                        ret["message"] = (
                            "Commit check passed, but skipping commit for dry-run and"
                            ' while rolling back configuration exception occurred "{}"'.format(
                                exception
                            )
                        )
                        ret["out"] = False
                        _restart_connection()

                try:
                    if write_diff and config_diff is not None:
                        with salt.utils.files.fopen(write_diff, "w") as fp:
                            fp.write(salt.utils.stringutils.to_str(config_diff))
                except Exception as exception:  # pylint: disable=broad-except
                    ret[
                        "message"
                    ] = "Could not write into diffs_file due to: '{}'".format(exception)
                    ret["out"] = False

        except ValueError as ex:
            message = "install_config failed due to: {}".format(str(ex))
            log.error(message)
            ret["message"] = message
            ret["out"] = False
        except LockError as ex:
            log.error("Configuration database is locked")
            ret["message"] = ex.message
            ret["out"] = False
        except RpcTimeoutError as ex:
            message = "install_config failed due to timeout error : {}".format(str(ex))
            log.error(message)
            ret["message"] = message
            ret["out"] = False
        except Exception as exc:  # pylint: disable=broad-except
            ret["message"] = "install_config failed due to exception: '{}'".format(exc)
            ret["out"] = False

        return ret


@_timeout_decorator_cleankwargs
def zeroize():
    """
    Resets the device to default factory settings

    .. note::
        In case of non-root user, proxy_reconnect will not be able
        to re-connect to the device as zeroize will delete the local
        user's configuration.
        For more details on zeroize functionality, please refer
        https://www.juniper.net/documentation/en_US/junos/topics/reference/command-summary/request-system-zeroize.html

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.zeroize
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    try:
        conn.cli("request system zeroize")
        ret["message"] = "Completed zeroize and rebooted"
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = 'Could not zeroize due to : "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()

    return ret


@_timeout_decorator
def install_os(path=None, **kwargs):
    """
    Installs the given image on the device. After the installation is complete
    the device is rebooted, if reboot=True is given as a keyworded argument.

    path (required)
        Path where the image file is present on the proxy minion

    remote_path : /var/tmp
        If the value of path  is a file path on the local
        (Salt host's) filesystem, then the image is copied from the local
        filesystem to the :remote_path: directory on the target Junos
        device. The default is ``/var/tmp``. If the value of :path: or
        is a URL, then the value of :remote_path: is unused.

    dev_timeout : 1800
        The NETCONF RPC timeout (in seconds). This argument was added since most of
        the time the "package add" RPC takes a significant amount of time.
        So this :timeout: value will be used in the context of the SW installation
        process.  Defaults to 30 minutes (30*60=1800 seconds)

    timeout : 1800
        Alias to dev_timeout for backward compatibility

    reboot : False
        Whether to reboot after installation

    no_copy : False
        If ``True`` the software package will not be SCP’d to the device

    bool validate:
        When ``True`` this method will perform a config validation against
        the new image

    bool issu: False
        When ``True`` allows unified in-service software upgrade
        (ISSU) feature enables you to upgrade between two different Junos OS
        releases with no disruption on the control plane and with minimal
        disruption of traffic.

    bool nssu: False
        When ``True`` allows nonstop software upgrade (NSSU)
        enables you to upgrade the software running on a Juniper Networks
        EX Series Virtual Chassis or a Juniper Networks EX Series Ethernet
        Switch with redundant Routing Engines with a single command and
        minimal disruption to network traffic.

    bool all_re: True
        When True (default), executes the software install on all Routing Engines of the Junos
        device. When False, execute the software install only on the current Routing Engine.

        .. versionadded:: 3001

    .. note::
        Any additional keyword arguments specified are passed down to PyEZ sw.install() as is.
        Please refer to below URl for PyEZ sw.install() documentation:
        https://pyez.readthedocs.io/en/latest/jnpr.junos.utils.html#jnpr.junos.utils.sw.SW.install

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.install_os 'salt://images/junos_image.tgz' reboot=True
        salt 'device_name' junos.install_os 'salt://junos_16_1.tgz' dev_timeout=300
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True

    op = {}
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    # timeout value is not honoured by sw.install if not passed as argument
    # currently, timeout is set to be maximum of default 1800 and user passed timeout value
    # For info: https://github.com/Juniper/salt/issues/116
    dev_timeout = max(op.pop("dev_timeout", 0), op.pop("timeout", 0))
    timeout = max(1800, conn.timeout, dev_timeout)

    # Reboot should not be passed as a keyword argument to install(),
    # Please refer to https://github.com/Juniper/salt/issues/115 for more details
    reboot = op.pop("reboot", False)
    no_copy_ = op.get("no_copy", False)

    if path is None:
        ret[
            "message"
        ] = "Please provide the salt path where the junos image is present."
        ret["out"] = False
        return ret

    if reboot:
        #  flag reboot active, disables proxy_reconnect since it's probing
        # of connection interferes with the reboot, especially with a
        # package install (time taken to xfer package and install)
        __proxy__["junos.reboot_active"]()

    install_status = False
    if not no_copy_:
        with HandleFileCopy(path) as image_path:
            if image_path is None:
                ret["message"] = "Invalid path. Please provide a valid image path"
                ret["out"] = False
                __proxy__["junos.reboot_clear"]()
                return ret
            if salt.utils.platform.is_junos():
                # If its native minion running on Junos, pyez dont need to SCP file
                # hence setting no_copy as True, HandleFileCopy already copied file
                # from master to Junos
                tmp_absfile = image_path
                op["no_copy"] = True
                op["remote_path"] = os.path.dirname(tmp_absfile)
                image_path = os.path.basename(tmp_absfile)
            try:
                install_status, install_message = conn.sw.install(
                    image_path, progress=True, timeout=timeout, **op
                )
            except Exception as exception:  # pylint: disable=broad-except
                ret["message"] = 'Installation failed due to: "{}"'.format(exception)
                ret["out"] = False
                __proxy__["junos.reboot_clear"]()
                _restart_connection()
                return ret
    else:
        try:
            install_status, install_message = conn.sw.install(
                path, progress=True, timeout=timeout, **op
            )
        except Exception as exception:  # pylint: disable=broad-except
            ret["message"] = 'Installation failed due to: "{}"'.format(exception)
            ret["out"] = False
            __proxy__["junos.reboot_clear"]()
            _restart_connection()
            return ret

    if install_status is True:
        ret["out"] = True
        ret["message"] = "Installed the os."
    else:
        ret["message"] = "Installation failed. Reason: {}".format(install_message)
        ret["out"] = False
        __proxy__["junos.reboot_clear"]()
        return ret

    # Handle reboot, after the install has finished
    if reboot is True:
        reboot_kwargs = {}
        if "vmhost" in op and op.get("vmhost") is True:
            reboot_kwargs["vmhost"] = True
        if "all_re" in op:
            reboot_kwargs["all_re"] = op.get("all_re")
        try:
            __proxy__["junos.reboot_active"]()
            conn.sw.reboot(**reboot_kwargs)
        except Exception as exception:  # pylint: disable=broad-except
            __proxy__["junos.reboot_clear"]()
            ret[
                "message"
            ] = 'Installation successful but reboot failed due to : "{}"'.format(
                exception
            )
            ret["out"] = False
            _restart_connection()
            return ret

        __proxy__["junos.reboot_clear"]()
        ret["out"] = True
        ret["message"] = "Successfully installed and rebooted!"

    return ret


@_timeout_decorator_cleankwargs
def file_copy(src, dest):
    """
    Copies the file from the local device to the junos device

    .. note::
        This function does not work on Juniper native minions

    src
        The source path where the file is kept.

    dest
        The destination path on the where the file will be copied

    .. versionadded:: 3001

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.file_copy /home/m2/info.txt info_copy.txt
    """
    if salt.utils.platform.is_junos():
        return {
            "success": False,
            "message": "This method is unsupported on the current operating system!",
        }

    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True

    with HandleFileCopy(src) as fp:
        if fp is None:
            ret["message"] = "Invalid source file path {}".format(src)
            ret["out"] = False
            return ret

        try:
            with SCP(conn, progress=True) as scp:
                scp.put(fp, dest)
            ret["message"] = "Successfully copied file from {} to {}".format(src, dest)
        except Exception as exception:  # pylint: disable=broad-except
            ret["message"] = 'Could not copy file : "{}"'.format(exception)
            ret["out"] = False

        return ret


@_timeout_decorator_cleankwargs
def lock():
    """
    Attempts an exclusive lock on the candidate configuration. This
    is a non-blocking call.

    .. note::
        When locking, it is important to remember to call
        :py:func:`junos.unlock <salt.modules.junos.unlock>` once finished. If
        locking during orchestration, remember to include a step in the
        orchestration job to unlock.

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.lock
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    try:
        conn.cu.lock()
        ret["message"] = "Successfully locked the configuration."
    except RpcTimeoutError as exception:
        ret["message"] = 'Could not gain lock due to : "{}"'.format(exception)
        ret["out"] = False
        _restart_connection()

    except LockError as exception:
        ret["message"] = 'Could not gain lock due to : "{}"'.format(exception)
        ret["out"] = False

    return ret


@_timeout_decorator_cleankwargs
def unlock():
    """
    Unlocks the candidate configuration.

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.unlock
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    try:
        conn.cu.unlock()
        ret["message"] = "Successfully unlocked the configuration."
    except RpcTimeoutError as exception:
        ret["message"] = 'Could not unlock configuration due to : "{}"'.format(
            exception
        )
        ret["out"] = False
        _restart_connection()

    except UnlockError as exception:
        ret["message"] = 'Could not unlock configuration due to : "{}"'.format(
            exception
        )
        ret["out"] = False

    return ret


@_timeout_decorator
def load(path=None, **kwargs):
    """
    Loads the configuration from the file provided onto the device.

    path (required)
        Path where the configuration/template file is present. If the file has
        a ``.conf`` extension, the content is treated as text format. If the
        file has a ``.xml`` extension, the content is treated as XML format. If
        the file has a ``.set`` extension, the content is treated as Junos OS
        ``set`` commands.

    overwrite : False
        Set to ``True`` if you want this file is to completely replace the
        configuration file. Sets action to override

        .. note:: This option cannot be used if **format** is "set".

    replace : False
        Specify whether the configuration file uses ``replace:`` statements. If
        ``True``, only those statements under the ``replace`` tag will be
        changed.

    merge : False
        If set to ``True`` will set the load-config action to merge.
        the default load-config action is 'replace' for xml/json/text config

    update : False
        Compare a complete loaded configuration against the candidate
        configuration. For each hierarchy level or configuration object that is
        different in the two configurations, the version in the loaded
        configuration replaces the version in the candidate configuration. When
        the configuration is later committed, only system processes that are
        affected by the changed configuration elements parse the new
        configuration. This action is supported from PyEZ 2.1.

    format
        Determines the format of the contents

    template_vars
      Variables to be passed into the template processing engine in addition to
      those present in pillar, the minion configuration, grains, etc.  You may
      reference these variables in your template like so:

      .. code-block:: jinja

          {{ template_vars["var_name"] }}

    CLI Examples:

    .. code-block:: bash

        salt 'device_name' junos.load 'salt://production/network/routers/config.set'

        salt 'device_name' junos.load 'salt://templates/replace_config.conf' replace=True

        salt 'device_name' junos.load 'salt://my_new_configuration.conf' overwrite=True

        salt 'device_name' junos.load 'salt://syslog_template.conf' template_vars='{"syslog_host": "10.180.222.7"}'
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True

    if path is None:
        ret[
            "message"
        ] = "Please provide the salt path where the configuration is present"
        ret["out"] = False
        return ret

    op = {}
    if "__pub_arg" in kwargs:
        if kwargs["__pub_arg"]:
            if isinstance(kwargs["__pub_arg"][-1], dict):
                op.update(kwargs["__pub_arg"][-1])
    else:
        op.update(kwargs)

    kwargs = {}
    if "template_vars" in op:
        kwargs.update({"template_vars": op["template_vars"]})

    with HandleFileCopy(path, **kwargs) as template_cached_path:
        if template_cached_path is None:
            ret["message"] = "Invalid file path."
            ret["out"] = False
            return ret

        if os.path.getsize(template_cached_path) == 0:
            ret["message"] = "Template failed to render"
            ret["out"] = False
            return ret

        op["path"] = template_cached_path

        if "format" not in op:
            if path.endswith("set"):
                template_format = "set"
            elif path.endswith("xml"):
                template_format = "xml"
            elif path.endswith("json"):
                template_format = "json"
            else:
                template_format = "text"

            op["format"] = template_format

        # Currently, four config_actions are supported: overwrite, replace, update, merge
        # Allow only one config_action, providing multiple config_action value is not allowed
        actions = [
            item
            for item in ("overwrite", "replace", "update", "merge")
            if op.get(item, False)
        ]
        if len(list(actions)) > 1:
            ret["message"] = "Only one config_action is allowed. Provided: {}".format(
                actions
            )
            ret["out"] = False
            return ret

        if "replace" in op and op["replace"]:
            op["merge"] = False
            del op["replace"]
        elif "overwrite" in op and op["overwrite"]:
            op["overwrite"] = True
        elif "merge" in op and op["merge"]:
            op["merge"] = True
        elif "overwrite" in op and not op["overwrite"]:
            op["merge"] = True
            del op["overwrite"]

        try:
            conn.cu.load(**op)
            ret["message"] = "Successfully loaded the configuration."
        except Exception as exception:  # pylint: disable=broad-except
            ret["message"] = 'Could not load configuration due to : "{}"'.format(
                exception
            )
            ret["format"] = op["format"]
            ret["out"] = False
            _restart_connection()
            return ret

        return ret


@_timeout_decorator_cleankwargs
def commit_check():
    """
    Perform a commit check on the configuration

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.commit_check
    """
    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    try:
        conn.cu.commit_check()
        ret["message"] = "Commit check succeeded."
    except Exception as exception:  # pylint: disable=broad-except
        ret["message"] = "Commit check failed with {}".format(exception)
        ret["out"] = False
        _restart_connection()

    return ret


@_timeout_decorator_cleankwargs
def get_table(
    table,
    table_file,
    path=None,
    target=None,
    key=None,
    key_items=None,
    filters=None,
    table_args=None,
):
    """
    .. versionadded:: 3001

    Retrieve data from a Junos device using Tables/Views

    table (required)
        Name of PyEZ Table

    table_file (required)
        YAML file that has the table specified in table parameter

    path:
        Path of location of the YAML file.
        defaults to op directory in jnpr.junos.op

    target:
        if command need to run on FPC, can specify fpc target

    key:
        To overwrite key provided in YAML

    key_items:
        To select only given key items

    filters:
        To select only filter for the dictionary from columns

    table_args:
        key/value pair which should render Jinja template command
        or are passed as args to rpc call in op table

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.get_table RouteTable routes.yml
        salt 'device_name' junos.get_table EthPortTable ethport.yml table_args='{"interface_name": "ge-3/2/2"}'
        salt 'device_name' junos.get_table EthPortTable ethport.yml salt://tables
    """

    conn = __proxy__["junos.conn"]()
    ret = {}
    ret["out"] = True
    ret["hostname"] = conn._hostname
    ret["tablename"] = table
    get_kvargs = {}
    if target is not None:
        get_kvargs["target"] = target
    if key is not None:
        get_kvargs["key"] = key
    if key_items is not None:
        get_kvargs["key_items"] = key_items
    if filters is not None:
        get_kvargs["filters"] = filters
    if table_args is not None and isinstance(table_args, dict):
        get_kvargs["args"] = table_args
    pyez_tables_path = os.path.dirname(os.path.abspath(tables_dir.__file__))
    try:
        if path is not None:
            file_path = os.path.join(path, "{}".format(table_file))
        else:
            file_path = os.path.join(pyez_tables_path, "{}".format(table_file))

        with HandleFileCopy(file_path) as file_loc:
            if file_loc is None:
                ret["message"] = "Given table file {} cannot be located".format(
                    table_file
                )
                ret["out"] = False
                return ret
            try:
                with salt.utils.files.fopen(file_loc) as fp:
                    ret["table"] = yaml.load(
                        fp.read(), Loader=yamlordereddictloader.Loader
                    )
                    globals().update(FactoryLoader().load(ret["table"]))
            except OSError as err:
                ret[
                    "message"
                ] = "Uncaught exception during YAML Load - please report: {}".format(
                    str(err)
                )
                ret["out"] = False
                return ret
            try:
                data = globals()[table](conn)
                data.get(**get_kvargs)
            except KeyError as err:
                ret[
                    "message"
                ] = "Uncaught exception during get API call - please report: {}".format(
                    str(err)
                )
                ret["out"] = False
                return ret
            except ConnectClosedError:
                ret[
                    "message"
                ] = "Got ConnectClosedError exception. Connection lost with {}".format(
                    conn
                )
                ret["out"] = False
                _restart_connection()
                return ret
            ret["reply"] = json.loads(data.to_json())
            if data.__class__.__bases__[0] in [OpTable, CfgTable]:
                # Sets key value if not present in YAML. To be used by returner
                if ret["table"][table].get("key") is None:
                    ret["table"][table]["key"] = data.ITEM_NAME_XPATH
                # If key is provided from salt state file.
                if key is not None:
                    ret["table"][table]["key"] = data.KEY
                if table_args is not None:
                    args = copy.copy(data.GET_ARGS)
                    args.update(table_args)
                    ret["table"][table]["args"] = args
            else:
                if target is not None:
                    ret["table"][table]["target"] = data.TARGET
                if key is not None:
                    ret["table"][table]["key"] = data.KEY
                if key_items is not None:
                    ret["table"][table]["key_items"] = data.KEY_ITEMS
                if table_args is not None:
                    args = copy.copy(data.CMD_ARGS)
                    args.update(table_args)
                    ret["table"][table]["args"] = args
                    ret["table"][table]["command"] = data.GET_CMD
    except ConnectClosedError:
        ret[
            "message"
        ] = "Got ConnectClosedError exception. Connection lost with {}".format(
            str(conn)
        )
        ret["out"] = False
        _restart_connection()
        return ret
    except Exception as err:  # pylint: disable=broad-except
        ret["message"] = "Uncaught exception - please report: {}".format(str(err))
        ret["out"] = False
        _restart_connection()
        return ret
    return ret


def _recursive_dict(node):
    """
    Convert an lxml.etree node tree into a dict.
    """
    result = {}

    for element in node.iterchildren():
        # Remove namespace prefix
        key = element.tag.split("}")[1] if "}" in element.tag else element.tag

        # Process element as tree element if the inner XML contains non-whitespace content
        if element.text and element.text.strip():
            value = element.text
        else:
            value = _recursive_dict(element)
        if key in result:

            if type(result[key]) is list:
                result[key].append(value)
            else:
                tempvalue = result[key].copy()
                result[key] = [tempvalue, value]
        else:
            result[key] = value
    return result


@_timeout_decorator
def rpc_file_list(path, **kwargs):
    """
    Use the Junos RPC interface to get a list of files and return
    them as a structure dictionary.

    .. versionadded:: 3003

    CLI Example:

    .. code-block:: bash

        salt junos-router junos.rpc_file_list /var/local/salt/etc

        junos-router:
            files:
                directory:
                    directory-name:
                        /var/local/salt/etc
                    file-information:
                        |_
                          file-directory:
                              file-name:
                                  pki
                        |_
                          file-name:
                              proxy
                        |_
                          file-directory:
                              file-name:
                                  proxy.d
                total-file-blocks:
                    10
                total-files:
                    1
        success:
            True

    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    conn = __proxy__["junos.conn"]()
    if conn._conn is None:
        return False

    results = conn.rpc.file_list(path=path)

    ret = {}
    ret["files"] = _recursive_dict(results)
    ret["success"] = True

    return ret


def _strip_newlines(str):
    stripped = str.replace("\n", "")
    return stripped


def _make_source_list(dir):

    dir_list = []
    if not dir:
        return
    base = rpc_file_list(dir)["files"]["directory"]

    # No files in this directory
    if "file-information" not in base:
        if "directory_name" not in base:
            return None
        return [os.path.join(_strip_newlines(base.get("directory-name", None))) + "/"]

    if isinstance(base["file-information"], dict):
        dirname = os.path.join(
            dir, _strip_newlines(base["file-information"]["file-name"])
        )
        if "file-directory" in base["file-information"]:
            new_list = _make_source_list(os.path.join(dir, dirname))
            return new_list
        else:
            return [dirname]
    for entry in base["file-information"]:
        if "file-directory" in entry:
            new_list = _make_source_list(
                os.path.join(dir, _strip_newlines(entry["file-name"]))
            )
            if new_list:
                dir_list.extend(new_list)
        else:
            dir_list.append(os.path.join(dir, _strip_newlines(entry["file-name"])))

    return dir_list


@_timeout_decorator
def file_compare(file1, file2, **kwargs):
    """
    Compare two files and return a dictionary indicating if they
    are different.

    Dictionary includes `success` key.  If False, one or more files do not
    exist or some other error occurred.

    Under the hood, this uses the junos CLI command `file compare files ...`

    .. note::
        This function only works on Juniper native minions

    .. versionadded:: 3003

    CLI Example:

    .. code-block:: bash

        salt junos-router junos.file_compare /var/tmp/backup1/cmt.script /var/tmp/backup2/cmt.script

        junos-router:
            identical:
                False
            success:
                True

    """
    if not salt.utils.platform.is_junos():
        return {
            "success": False,
            "message": "This method is unsupported on the current operating system!",
        }

    ret = {"message": "", "identical": False, "success": True}

    junos_cli = salt.utils.path.which("cli")
    if not junos_cli:
        return {"success": False, "message": "Cannot find Junos cli command"}

    cliret = __salt__["cmd.run"](
        "{} file compare files {} {} ".format(junos_cli, file1, file2)
    )
    clilines = cliret.splitlines()

    for r in clilines:
        if r.strip() != "":
            if "No such file" in r:
                ret["identical"] = False
                ret["success"] = False
                return ret

            ret["identical"] = False
            ret["success"] = True
            return ret

    ret["identical"] = True
    ret["success"] = True
    return ret


@_timeout_decorator
def fsentry_exists(dir, **kwargs):
    """
    Returns a dictionary indicating if `dir` refers to a file
    or a non-file (generally a directory) in the file system,
    or if there is no file by that name.

    .. note::
        This function only works on Juniper native minions

    .. versionadded:: 3003

    CLI Example:

    .. code-block:: bash

        salt junos-router junos.fsentry_exists /var/log

        junos-router:
            is_dir:
                True
            exists:
                True

    """
    if not salt.utils.platform.is_junos():
        return {
            "success": False,
            "message": "This method is unsupported on the current operating system!",
        }

    junos_cli = salt.utils.path.which("cli")
    if not junos_cli:
        return {"success": False, "message": "Cannot find Junos cli command"}

    ret = __salt__["cmd.run"]("{} file show {}".format(junos_cli, dir))
    retlines = ret.splitlines()
    exists = True
    is_dir = False
    status = {"is_dir": False, "exists": True}
    for r in retlines:
        if "could not resolve" in r or "error: Could not connect" in r:
            status["is_dir"] = False
            status["exists"] = False
        if "is not a regular file" in r:
            status["is_dir"] = True
            status["exists"] = True

    return status


def _find_routing_engines():
    junos_cli = salt.utils.path.which("cli")
    if not junos_cli:
        return {"success": False, "message": "Cannot find Junos cli command"}

    re_check = __salt__["cmd.run"]("{} show chassis routing-engine".format(junos_cli))
    engine_present = True
    engine = {}

    # for l in re_check.splitlines():
    #     if 'error: Unrecognized command' in l:
    #         engine_present = False

    # if not engine_present:
    #     return {'success': False,
    #             'message': 'Device does not have multiple routing engines'}

    current_engine = None
    status = None
    for l in re_check.splitlines():
        if "Slot" in l:
            mat = re.search(".*(\\d+):.*", l)
            if mat:
                current_engine = "re" + str(mat.group(1)) + ":"
        if "Current state" in l:
            if "Master" in l:
                status = "Master"
            if "Disabled" in l:
                status = "Disabled"
            if "Backup" in l:
                status = "Backup"

        if current_engine and status:
            engine[current_engine] = status
            current_engine = None
            status = None

    if not engine:
        return {
            "success": False,
            "message": "Junos cli command returned no information",
        }

    engine["success"] = True
    return engine


@_timeout_decorator
def routing_engine(**kwargs):
    """
    Returns a dictionary containing the routing engines on the device and
    their status (Master, Disabled, Backup).

    Under the hood parses the result of `show chassis routing-engine`

    .. versionadded:: 3003

    CLI Example:

    .. code-block:: bash

        salt junos-router junos.routing_engine

        junos-router:
            backup:
              - re1:
            master:
              re0:
            success:
              True

    Returns `success: False` if the device does not appear to have multiple routing engines.

    """
    engine_status = _find_routing_engines()
    if not engine_status["success"]:
        return {"success": False}

    master = None
    backup = []
    for k, v in engine_status.items():
        if v == "Master":
            master = k
        if v == "Backup" or v == "Disabled":
            backup.append(k)

    if master:
        ret = {"master": master, "backup": backup, "success": True}
    else:
        ret = {"master": master, "backup": backup, "success": False}
    log.debug(ret)
    return ret


@_timeout_decorator
def dir_copy(source, dest, force=False, **kwargs):
    """
    Copy a directory and recursively its contents from source to dest.

    .. note::
        This function only works on the Juniper native minion

    Parameters:

    source : Directory to use as the source

    dest : Directory in which to place the source and its contents.

    force : This function will not copy identical files unless `force` is `True`

    .. versionadded:: 3003

    CLI Example:

    .. code-block:: bash

        salt 'device_name' junos.dir_copy /etc/salt/pki re1:/

    This will take the `pki` directory, its absolute path and copy it and its
    contents to routing engine 1 root directory. The result will be
    `re1:/etc/salt/pki/<files and dirs in /etc/salt/pki`.

    """
    if not salt.utils.platform.is_junos():
        return {
            "success": False,
            "message": "This method is unsupported on the current operating system!",
        }

    junos_cli = salt.utils.path.which("cli")
    if not junos_cli:
        return {"success": False, "message": "Cannot find Junos cli command"}

    ret = {}
    ret_messages = ""
    if not source.startswith("/"):
        ret["message"] = "Source directory must be a fully qualified path."
        ret["success"] = False
        return ret

    if not (dest.endswith(":") or dest.startswith("/")):
        ret["message"] = (
            "Destination must be a routing engine reference (e.g. re1:) or a fully"
            " qualified path."
        )
        ret["success"] = False
        return ret

    check_source = fsentry_exists(source)
    if not check_source["exists"]:
        ret["message"] = "Source does not exist"
        ret["success"] = False
        return ret

    if not check_source["is_dir"]:
        ret["message"] = "Source is not a directory."
        ret["success"] = False
        return ret

    filelist = _make_source_list(source)
    dirops = []
    for f in filelist:
        splitpath = os.path.split(f)[0]
        fullpath = "/"
        for component in splitpath.split("/"):
            fullpath = os.path.join(fullpath, component)
            if fullpath not in dirops:
                dirops.append(fullpath)

    for d in dirops:
        target = dest + d
        status = fsentry_exists(target)
        if not status["exists"]:
            ret = __salt__["cmd.run"](
                "{} file make-directory {}".format(junos_cli, target)
            )
            ret = ret_messages + ret
        else:
            ret_messages = ret_messages + "Directory " + target + " already exists.\n"
    for f in filelist:
        if not f.endswith("/"):
            target = dest + f
            comp_result = file_compare(f, target)

            if not comp_result["identical"] or force:
                ret = __salt__["cmd.run"](
                    "{} file copy {} {}".format(junos_cli, f, target)
                )
                ret = ret_messages + ret
            else:
                ret_messages = (
                    ret_messages
                    + "Files {} and {} are identical, not copying.\n".format(f, target)
                )

    return ret_messages



"""
Integration tests for the etcd modules
"""

import logging

import pytest

from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

etcd = pytest.importorskip("etcd")


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]


@pytest.fixture(scope="module")
def etcd_static_port(sdb_etcd_port):  # pylint: disable=function-redefined
    return sdb_etcd_port


@pytest.fixture(
    scope="module",
    params=(EtcdVersion.v2, EtcdVersion.v3_v2_mode),
    ids=etcd_version_ids,
)  # pylint: disable=function-redefined
def etcd_version(request):
    # The only parameter is True because the salt integration
    # configuration for the salt-master and salt-minion defaults
    # to v2.
    # The alternative would be to start a separate master and minion
    # to try both versions.
    # Maybe at another time, though, we're deprecating version 2, so,
    # it might not be worth it.
    if request.param == EtcdVersion.v2 and not HAS_ETCD_V2:
        pytest.skip("No etcd library installed")
    if request.param != EtcdVersion.v2 and not HAS_ETCD_V3:
        pytest.skip("No etcd3 library installed")
    return request.param


@pytest.fixture(scope="module", autouse=True)
def _container_running(etcd_container):
    return etcd_container


def test_sdb(salt_call_cli):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbetcd/secret/test/test_sdb/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "bar"

    ret = salt_call_cli.run("sdb.get", uri="sdb://sdbetcd/secret/test/test_sdb/foo")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "bar"


def test_sdb_runner(salt_run_cli):
    ret = salt_run_cli.run(
        "sdb.set", uri="sdb://sdbetcd/secret/test/test_sdb_runner/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.stdout == "bar"

    ret = salt_run_cli.run(
        "sdb.get", uri="sdb://sdbetcd/secret/test/test_sdb_runner/foo"
    )
    assert ret.returncode == 0
    assert ret.stdout == "bar"


def test_config(salt_call_cli, pillar_tree):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbetcd/secret/test/test_pillar_sdb/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "bar"

    ret = salt_call_cli.run("config.get", "test_etcd_pillar_sdb")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "bar"



"""
Create/Close an alert in OpsGenie
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 2018.3.0

This state is useful for creating or closing alerts in OpsGenie
during state runs.

.. code-block:: yaml

    used_space:
      disk.status:
        - name: /
        - maximum: 79%
        - minimum: 20%

    opsgenie_create_action_sender:
      opsgenie.create_alert:
        - api_key: XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
        - reason: 'Disk capacity is out of designated range.'
        - name: disk.status
        - onfail:
          - disk: used_space

    opsgenie_close_action_sender:
      opsgenie.close_alert:
        - api_key: XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
        - name: disk.status
        - require:
          - disk: used_space

"""

import inspect
import logging

import salt.exceptions

log = logging.getLogger(__name__)


def create_alert(name=None, api_key=None, reason=None, action_type="Create"):
    """
    Create an alert in OpsGenie. Example usage with Salt's requisites and other
    global state arguments could be found above.

    Required Parameters:

    api_key
        It's the API Key you've copied while adding integration in OpsGenie.

    reason
        It will be used as alert's default message in OpsGenie.

    Optional Parameters:

    name
        It will be used as alert's alias. If you want to use the close
        functionality you must provide name field for both states like
        in above case.

    action_type
        OpsGenie supports the default values Create/Close for action_type.
        You can customize this field with OpsGenie's custom actions for
        other purposes like adding notes or acknowledging alerts.
    """

    _, _, _, values = inspect.getargvalues(inspect.currentframe())
    log.info("Arguments values: %s", values)

    ret = {"result": "", "name": "", "changes": "", "comment": ""}

    if api_key is None or reason is None:
        raise salt.exceptions.SaltInvocationError("API Key or Reason cannot be None.")

    if __opts__["test"] is True:
        ret[
            "comment"
        ] = 'Test: {} alert request will be processed using the API Key="{}".'.format(
            action_type, api_key
        )

        # Return ``None`` when running with ``test=true``.
        ret["result"] = None

        return ret

    response_status_code, response_text = __salt__["opsgenie.post_data"](
        api_key=api_key, name=name, reason=reason, action_type=action_type
    )

    if 200 <= response_status_code < 300:
        log.info(
            "POST Request has succeeded with message: %s status code: %s",
            response_text,
            response_status_code,
        )
        ret[
            "comment"
        ] = 'Test: {} alert request will be processed using the API Key="{}".'.format(
            action_type, api_key
        )
        ret["result"] = True
    else:
        log.error(
            "POST Request has failed with error: %s status code: %s",
            response_text,
            response_status_code,
        )
        ret["result"] = False

    return ret


def close_alert(
    name=None, api_key=None, reason="Conditions are met.", action_type="Close"
):
    """
    Close an alert in OpsGenie. It's a wrapper function for create_alert.
    Example usage with Salt's requisites and other global state arguments
    could be found above.

    Required Parameters:

    name
        It will be used as alert's alias. If you want to use the close
        functionality you must provide name field for both states like
        in above case.

    Optional Parameters:

    api_key
        It's the API Key you've copied while adding integration in OpsGenie.

    reason
        It will be used as alert's default message in OpsGenie.

    action_type
        OpsGenie supports the default values Create/Close for action_type.
        You can customize this field with OpsGenie's custom actions for
        other purposes like adding notes or acknowledging alerts.
    """
    if name is None:
        raise salt.exceptions.SaltInvocationError("Name cannot be None.")

    return create_alert(name, api_key, reason, action_type)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.introspect
"""


import pytest

import salt.modules.introspect as introspect
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {introspect: {}}


# 'running_service_owners' function tests: 1


def test_running_service_owners():
    """
    Test if it determine which packages own the currently running services.
    """
    err1 = (
        "The module for the package manager on this system does not"
        " support looking up which package(s) owns which file(s)"
    )
    err2 = (
        "The file module on this system does not "
        "support looking up open files on the system"
    )
    ret = {
        "Error": {
            "Unsupported File Module": "{}".format(err2),
            "Unsupported Package Manager": "{}".format(err1),
        }
    }
    assert introspect.running_service_owners() == ret

    mock = MagicMock(return_value={})
    with patch.dict(
        introspect.__salt__,
        {"pkg.owner": mock, "file.open_files": mock, "service.execs": mock},
    ):
        assert introspect.running_service_owners() == {}


# 'enabled_service_owners' function tests: 1


def test_enabled_service_owners():
    """
    Test if it return which packages own each of the services
    that are currently enabled.
    """
    err1 = (
        "The module for the package manager on this system does not"
        " support looking up which package(s) owns which file(s)"
    )
    err2 = (
        "The module for the service manager on this system does not"
        " support showing descriptive service data"
    )
    ret = {
        "Error": {
            "Unsupported Service Manager": "{}".format(err2),
            "Unsupported Package Manager": "{}".format(err1),
        }
    }
    assert introspect.enabled_service_owners() == ret

    mock = MagicMock(return_value={})
    with patch.dict(
        introspect.__salt__,
        {"pkg.owner": mock, "service.show": mock, "service.get_enabled": mock},
    ):
        assert introspect.enabled_service_owners() == {}


# 'service_highstate' function tests: 1


def test_service_highstate():
    """
    Test if it return running and enabled services in a highstate structure.
    """
    with patch(
        "salt.modules.introspect.running_service_owners", MagicMock(return_value={})
    ), patch(
        "salt.modules.introspect.enabled_service_owners", MagicMock(return_value={})
    ):
        assert introspect.service_highstate() == {}



"""
Library for interacting with Mattermost Incoming Webhooks
:configuration: This module can be used by specifying the name of a
    configuration profile in the minion config, minion pillar, or master
    config.
    For example:
    .. code-block:: yaml

        mattermost:
          hook: 3tdgo8restnxiykdx88wqtxryr
          api_url: https://example.com
"""

import http.client
import logging
import urllib.parse

import salt.utils.http
from salt.version import __version__

log = logging.getLogger(__name__)


def query(hook=None, api_url=None, data=None):
    """
    Mattermost object method function to construct and execute on the API URL.
    :param api_url:     The Mattermost API URL
    :param hook:        The Mattermost hook.
    :param data:        The data to be sent for POST method.
    :return:            The json response from the API call or False.
    """
    method = "POST"

    ret = {"message": "", "res": True}

    base_url = urllib.parse.urljoin(api_url, "/hooks/")
    url = urllib.parse.urljoin(base_url, str(hook))

    result = salt.utils.http.query(url, method, data=data, decode=True, status=True)

    if result.get("status", None) == http.client.OK:
        ret["message"] = "Message posted {} correctly".format(data)
        return ret
    elif result.get("status", None) == http.client.NO_CONTENT:
        return True
    else:
        log.debug(url)
        log.debug(data)
        log.debug(result)
        if "dict" in result:
            _result = result["dict"]
            if "error" in _result:
                ret["message"] = result["error"]
                ret["res"] = False
                return ret
            ret["message"] = "Message not posted"
        else:
            ret["message"] = "invalid_auth"
            ret["res"] = False
    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import pytest

import salt.modules.useradd as useradd
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


class UserAddTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.useradd
    """

    def setup_loader_modules(self):
        return {useradd: {}}

    @classmethod
    def setUpClass(cls):
        cls.mock_pwall = {
            "gid": 0,
            "groups": ["root"],
            "home": "/root",
            "name": "root",
            "passwd": "x",
            "shell": "/bin/bash",
            "uid": 0,
            "fullname": "root",
            "roomnumber": "",
            "workphone": "",
            "homephone": "",
            "other": "",
        }

    @classmethod
    def tearDownClass(cls):
        del cls.mock_pwall

    # 'getent' function tests: 2

    @pytest.mark.skipif(HAS_PWD is False, reason="The pwd module is not available")
    def test_getent(self):
        """
        Test if user.getent already have a value
        """
        with patch("salt.modules.useradd.__context__", MagicMock(return_value="Salt")):
            self.assertTrue(useradd.getent())

    @pytest.mark.skipif(HAS_PWD is False, reason="The pwd module is not available")
    def test_getent_user(self):
        """
        Tests the return information on all users
        """
        with patch("pwd.getpwall", MagicMock(return_value=[""])):
            ret = [
                {
                    "gid": 0,
                    "groups": ["root"],
                    "home": "/root",
                    "name": "root",
                    "passwd": "x",
                    "shell": "/bin/bash",
                    "uid": 0,
                    "fullname": "root",
                    "roomnumber": "",
                    "workphone": "",
                    "homephone": "",
                    "other": "",
                }
            ]
            with patch(
                "salt.modules.useradd._format_info",
                MagicMock(return_value=self.mock_pwall),
            ):
                self.assertEqual(useradd.getent(), ret)

    # 'info' function tests: 1

    @pytest.mark.skipif(HAS_PWD is False, reason="The pwd module is not available")
    def test_info(self):
        """
        Test the user information
        """
        self.assertEqual(useradd.info("username-that-does-not-exist"), {})

        mock = MagicMock(
            return_value=pwd.struct_passwd(
                (
                    "_TEST_GROUP",
                    "*",
                    83,
                    83,
                    "AMaViS Daemon",
                    "/var/virusmails",
                    "/usr/bin/false",
                )
            )
        )
        with patch.object(pwd, "getpwnam", mock):
            self.assertEqual(
                useradd.info("username-that-does-not-exist")["name"], "_TEST_GROUP"
            )

    # 'list_groups' function tests: 1

    def test_list_groups(self):
        """
        Test if it return a list of groups the named user belongs to
        """
        with patch("salt.utils.user.get_group_list", MagicMock(return_value="Salt")):
            self.assertEqual(useradd.list_groups("name"), "Salt")

    # 'list_users' function tests: 1

    @pytest.mark.skipif(HAS_PWD is False, reason="The pwd module is not available")
    def test_list_users(self):
        """
        Test if it returns a list of all users
        """
        self.assertTrue(useradd.list_users())

    def test_build_gecos_field(self):
        """
        Test if gecos fields are built correctly (removing trailing commas)
        """
        test_gecos = {
            "fullname": "Testing",
            "roomnumber": 1234,
            "workphone": 22222,
            "homephone": 99999,
        }
        expected_gecos_fields = "Testing,1234,22222,99999"
        self.assertEqual(useradd._build_gecos(test_gecos), expected_gecos_fields)
        test_gecos.pop("roomnumber")
        test_gecos.pop("workphone")
        expected_gecos_fields = "Testing,,,99999"
        self.assertEqual(useradd._build_gecos(test_gecos), expected_gecos_fields)
        test_gecos.pop("homephone")
        expected_gecos_fields = "Testing"
        self.assertEqual(useradd._build_gecos(test_gecos), expected_gecos_fields)



"""
Wrapper function for mine operations for salt-ssh

.. versionadded:: 2015.5.0
"""


import copy

import salt.client.ssh


def get(tgt, fun, tgt_type="glob", roster="flat"):
    """
    Get data from the mine based on the target, function and tgt_type

    This will actually run the function on all targeted minions (like
    publish.publish), as salt-ssh clients can't update the mine themselves.

    We will look for mine_functions in the roster, pillar, and master config,
    in that order, looking for a match for the defined function

    Targets can be matched based on any standard matching system that can be
    matched on the defined roster (in salt-ssh) via these keywords::

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' mine.get '*' network.interfaces
        salt-ssh '*' mine.get 'myminion' network.interfaces roster=flat
        salt-ssh '*' mine.get '192.168.5.0' network.ipaddrs roster=scan
    """
    # Set up opts for the SSH object
    opts = copy.deepcopy(__context__["master_opts"])
    minopts = copy.deepcopy(__opts__)
    opts.update(minopts)
    if roster:
        opts["roster"] = roster
    opts["argv"] = [fun]
    opts["selected_target_option"] = tgt_type
    opts["tgt"] = tgt
    opts["arg"] = []

    # Create the SSH object to handle the actual call
    ssh = salt.client.ssh.SSH(opts)

    # Run salt-ssh to get the minion returns
    rets = {}
    for ret in ssh.run_iter(mine=True):
        rets.update(ret)

    cret = {}
    for host in rets:
        if "return" in rets[host]:
            cret[host] = rets[host]["return"]
        else:
            cret[host] = rets[host]
    return cret



"""test for pillar csvpillar.py"""


import salt.pillar.csvpillar as csvpillar
from tests.support.mock import mock_open, patch


def test_001_load_utf8_csv():
    fake_csv = "id,foo,bar\r\nminion1,foo1,bar1"
    fake_dict = {"id": "minion1", "foo": "foo1", "bar": "bar1"}
    fopen_mock = mock_open(fake_csv)
    with patch("salt.utils.files.fopen", fopen_mock):
        result = csvpillar.ext_pillar(
            mid="minion1",
            pillar=None,
            path="/fake/path/file.csv",
            idkey="id",
            namespace=None,
        )
        assert fake_dict == result


def test_002_load_utf8_csv_namespc():
    fake_csv = "id,foo,bar\r\nminion1,foo1,bar1"
    fake_dict = {"baz": {"id": "minion1", "foo": "foo1", "bar": "bar1"}}
    fopen_mock = mock_open(fake_csv)
    with patch("salt.utils.files.fopen", fopen_mock):
        result = csvpillar.ext_pillar(
            mid="minion1",
            pillar=None,
            path="/fake/path/file.csv",
            idkey="id",
            namespace="baz",
        )
        assert fake_dict == result
